import pathmagic  # noqa
from torch.utils.data.dataset import Dataset
from scipy.spatial.transform import Rotation
from data.data_parser import UnityRoom
import data.visualizer as visualizer
import data.variables as variables
import numpy as np
import cv2 as cv
import math
import vedo
import os


os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"
PROJECT_DIR = os.path.dirname(os.path.dirname(__file__))
EPS = 1e-3


class ThreedFrontDataset(Dataset):
    def __init__(self, room_type='all', split='train', rot_aug=False, group=False, as_user_intent=False):
        ''' room_type: all, bedroom, diningroom, library, livingroom
        '''
        # folders
        self.data_root = os.path.join(pathmagic.PROJECT_DIR, 'dataset')
        self.data_folder = '3D-FRONT-labeled'
        self.room_type = room_type
        self.data_jsons = []
        self.data_floormasks = []
        train_room_json, test_room_json, train_floor_mask, test_floor_mask = self.load_json(room_type)
        if split == 'train':
            self.data_jsons = train_room_json
            self.data_floormasks = train_floor_mask
        elif split == 'test':
            self.data_jsons = test_room_json
            self.data_floormasks = test_floor_mask
        elif split == 'traintest':
            self.data_jsons = train_room_json + test_room_json
            self.data_floormasks = train_floor_mask + test_floor_mask
        self.rot_aug = rot_aug
        self.group = group
        self.as_user_intent = as_user_intent
        self.nclass = 27
        self.nclass_group = 12

    def __len__(self):
        return len(self.data_jsons)

    def __getitem__(self, idx):
        _, pos, rot, scale, cates, obj_idx, group_inst, group_lab, floor_mask, floor_scales = self.load_room(idx)
        if self.rot_aug:
            np.random.seed()
            times = np.random.randint(low=0, high=4)
            pos, rot = self.rotation_augmentation(pos, rot, angle=90*times)
        # pos:3, rot:1, scale:3, cates:1, obj_idx:1. total:9
        data = np.hstack([pos, rot.reshape(-1, 1), scale, cates.reshape(-1, 1), obj_idx.reshape(-1, 1)])
        group_bbox_list = self.parse_groups(data, group_inst)
        if not self.group:
            group_posrot = self.parse_group_posrot(group_bbox_list)
            if self.rot_aug:
                group_posrot = self.normalize_group(group_posrot, floor_scales, times)
            else:
                group_posrot = self.normalize_group(group_posrot, floor_scales)
            group_onehot = self.parse_group_onehot(group_lab)
            N = group_posrot.shape[0]
            group_posrot, group_onehot, group_lab = self.padding_group_data(group_posrot, group_onehot, group_lab, 10)
            if self.as_user_intent:
                complexity = []
                for i in range(len(group_bbox_list)):
                    complexity.append(group_bbox_list[i].shape[0])
                complexity = np.array(complexity)
                return group_posrot, group_onehot, group_lab, floor_mask, N, floor_scales, complexity
            else:
                return group_posrot, group_onehot, group_lab, floor_mask, N, floor_scales
        else:
            for i in range(len(group_bbox_list)):
                group_bbox_list[i][:, 8] = i
            all_bbox = np.vstack(group_bbox_list)
            bbox_onehot = np.eye(self.nclass)[all_bbox[:, 7].astype('int16')]
            return all_bbox, bbox_onehot, group_lab

    def parse_group_posrot(self, group_bbox_list):
        ''' add start/end token
        '''
        posrot_all = []
        for i in range(len(group_bbox_list)):
            pos = group_bbox_list[i][:, 0:3].mean(axis=0).reshape(1, -1)
            scale = group_bbox_list[i][:, 4] * group_bbox_list[i][:, 5] * group_bbox_list[i][:, 6]
            idx = scale.argmax()
            rot = group_bbox_list[i][idx, 3].reshape(1, 1)
            posrot = np.hstack([pos, rot])
            posrot_all.append(posrot)
        posrot_all = np.vstack(posrot_all)
        return posrot_all

    def parse_group_onehot(self, group_lab):
        ''' add start/end token
        '''
        return np.eye(self.nclass_group)[group_lab]

    def parse_groups(self, bbox, group_inst):
        group_list = []
        group_bbox_list = []
        for i in range(len(group_inst)):
            group = []
            for j in range(len(group_inst[i])):
                for k in range(len(bbox)):
                    if group_inst[i][j] == bbox[k, 8]:
                        group.append(bbox[k].reshape(1, -1))
            group_list.append(np.vstack(group))
        for i in range(len(group_list)):
            group_bbox_list.append(group_list[i][:, :9])
        return group_bbox_list

    def padding_group_data(self, group_posrot, group_onehot, group_lab, N_max):
        N = group_posrot.shape[0]
        group_posrot = np.vstack([group_posrot, np.zeros((N_max-N, 4))])
        group_onehot = np.vstack([group_onehot, np.zeros((N_max-N, self.nclass_group))])
        group_lab = np.hstack([group_lab, np.zeros(N_max-N)])
        return group_posrot, group_onehot, group_lab

    def normalize_group(self, group_posrot, floor_scales, times=0):
        ''' pos [-4, 4]->[-1, 1]; rot [0, 360]->[-1, 1]
        '''
        group_posrot[:, 0:3] = group_posrot[:, 0:3] / 4
        if times == 0 or times == 2:
            group_posrot[:, 0] = group_posrot[:, 0] / floor_scales[1]
            group_posrot[:, 2] = group_posrot[:, 2] / floor_scales[0]
        elif times == 1 or times == 3:
            group_posrot[:, 0] = group_posrot[:, 0] / floor_scales[0]
            group_posrot[:, 2] = group_posrot[:, 2] / floor_scales[1]
        group_posrot[:, 3] = ((group_posrot[:, 3] % 360) - 180.0) / 180.0
        group_posrot = np.clip(group_posrot, a_min=-1, a_max=1)
        return group_posrot

    def rotation_augmentation(self, pos, rot, angle=0):
        rotMatrix = Rotation.from_rotvec((angle / 360.) * 2 * math.pi * np.array([0, 1, 0])).as_matrix()
        R = np.array(rotMatrix)
        pos = np.transpose(pos)
        pos = np.matmul(R, pos)
        pos = np.transpose(pos)
        rot = rot + angle
        return pos, rot

    def load_room(self, idx):
        # print(self.data_jsons[idx])
        data_json = os.path.join(self.data_root, self.data_folder, self.room_type, self.data_jsons[idx])
        data_floormask = os.path.join(self.data_root, self.data_folder, self.room_type, self.data_floormasks[idx])
        room = UnityRoom(data_json, data_floormask, variables.FUTURE_FOLDER, variables.FLOOR_FOLDER)
        pos, rot, scale, cates, obj_idx, group_inst, group_lab = room.parse_numpy_input_gt()
        group_inst = self.ensure_group_index_valid(obj_idx, group_inst)
        floor_mask, floor_scales = self.process_floor_mask(data_floormask)
        return room, pos, rot, scale, cates, obj_idx, group_inst, group_lab, floor_mask, floor_scales

    def load_json(self, room_type='bedroom'):
        train_all_room_json, test_all_room_json = [], []
        train_all_floor_mask, test_all_floor_mask = [], []
        type_list = ['bedroom', 'diningroom', 'library', 'livingroom']
        if room_type == 'all':
            for type in type_list:
                train_room_json, test_room_json, train_floor_mask, test_floor_mask = self.read_txt(type)
                train_all_room_json += train_room_json
                test_all_room_json += test_room_json
                train_all_floor_mask += train_floor_mask
                test_all_floor_mask += test_floor_mask
            set_train_all_room_json = set(train_all_room_json)
            set_test_all_room_json = set(test_all_room_json)
            set_train_all_floor_mask = set(train_all_floor_mask)
            set_test_all_floor_mask = set(test_all_floor_mask)
            intersection_room = set_train_all_room_json.intersection(set_test_all_room_json)
            intersection_floor = set_train_all_floor_mask.intersection(set_test_all_floor_mask)
            train_all_room_json = list(set_train_all_room_json - intersection_room)
            train_all_floor_mask = list(set_train_all_floor_mask - intersection_floor)
            test_all_room_json = list(set_test_all_room_json)
            test_all_floor_mask = list(set_test_all_floor_mask)
            return train_all_room_json, test_all_room_json, train_all_floor_mask, test_all_floor_mask
        else:
            train_room_json, test_room_json, train_floor_mask, test_floor_mask = self.read_txt(room_type)
        return train_room_json, test_room_json, train_floor_mask, test_floor_mask

    def read_txt(self, room_type='bedroom'):
        train_room_json, test_room_json, train_floor_mask, test_floor_mask = [], [], [], []
        with open(os.path.join(self.data_root, f'train_room_{room_type}.txt'), 'r') as f:
            lines = f.readlines()
        f.close()
        for i in range(len(lines)):
            train_room_json.append(lines[i].strip())

        with open(os.path.join(self.data_root, f'test_room_{room_type}.txt'), 'r') as f:
            lines = f.readlines()
        f.close()
        for i in range(len(lines)):
            test_room_json.append(lines[i].strip())

        with open(os.path.join(self.data_root, f'train_floor_{room_type}.txt'), 'r') as f:
            lines = f.readlines()
        f.close()
        for i in range(len(lines)):
            train_floor_mask.append(lines[i].strip())

        with open(os.path.join(self.data_root, f'test_floor_{room_type}.txt'), 'r') as f:
            lines = f.readlines()
        f.close()
        for i in range(len(lines)):
            test_floor_mask.append(lines[i].strip())

        return train_room_json, test_room_json, train_floor_mask, test_floor_mask

    def ensure_group_index_valid(self, obj_idx, group_inst):
        for group in group_inst:
            for idx in group:
                if idx not in obj_idx:
                    group.remove(idx)
        return group_inst

    def process_floor_mask(self, data_floormask):
        img = cv.imread(data_floormask, cv.IMREAD_GRAYSCALE)
        # cv.imshow('test', img)
        # cv.waitKey()
        img = cv.resize(img, (64, 64))
        valid_idx = []
        for i in range(img.shape[0]):
            for j in range(img.shape[1]):
                if img[i, j] != 0:
                    img[i, j] = 255
                    valid_idx.append([i, j])
        valid_idx = np.array(valid_idx)
        maxs = np.max(valid_idx, axis=0)
        mins = np.min(valid_idx, axis=0)
        floor_scales = (maxs - mins) / 64
        return img, floor_scales

    def collect_model_pool(self, idx, collect_dict: dict):
        room, _, _, _, cates, _, _, _, _, _ = self.load_room(idx)
        for i in range(len(cates)):
            if str(cates[i]) not in collect_dict.keys():
                collect_dict[str(cates[i])] = [room.instance_jids[i]]
            else:
                collect_dict[str(cates[i])].append(room.instance_jids[i])
        return collect_dict

    def visualize_scene(self, idx, show_mesh=True, return_vis=False):
        ''' pos, translation (N, 3)
            rot, angle (N): 360 angle % 360
            scale, size (N, 3)
        '''
        room, pos, rot, scale, _, _, _, _, _, _ = self.load_room(idx)
        # pos, rot, _ = self.rotation_augmentation(pos, rot, None, angle=45)
        vedo_vis = []
        light = vedo.Light(pos=[0, 4, 0], c='white', intensity=1.2)
        for i in range(len(pos)):
            # bounding box
            unit_bbox = visualizer.unit_bbox_generator(r=10)
            bbox, arrow = visualizer.transform_points(unit_bbox, pos[i], rot[i], scale[i])
            lines = visualizer.bbox_to_lines(bbox, lw=5)
            arrow = vedo.Arrow(pos[i], pos[i]+arrow)
            vedo_vis.append(bbox)
            vedo_vis.append(lines)
            vedo_vis.append(arrow)
            if show_mesh:
                # mesh
                mesh_file = os.path.join(room.future_folder, room.instance_jids[i], 'raw_model.obj')
                tex_file = os.path.join(room.future_folder, room.instance_jids[i], 'texture.png')
                tex = cv.imread(tex_file)
                mesh_obj = vedo.Mesh(mesh_file)
                mesh_obj.texture(tex)
                instance_mesh, _ = room.unity_transform_mesh(mesh_obj, room.positions[i], room.rotations[i], room.scales[i])
                instance_mesh.alpha(0.9)
                vedo_vis.append(instance_mesh)
        floor_mesh = visualizer.load_floor(room, os.path.join(PROJECT_DIR, 'floor_texture'), 3)
        vedo_vis.append(floor_mesh)
        if return_vis:
            return vedo_vis, floor_mesh, light
        else:
            vedo.show(vedo_vis, light, axes=2, interactive=1)

    def visualize_floor_mesh(self, idx, floor):
        ''' pos, translation (N, 3)
            rot, angle (N): 360 angle % 360
            scale, size (N, 3)
        '''
        room, pos, rot, scale, _, _, _, _, _, _ = self.load_room(idx)
        floor_mesh = visualizer.load_floor(room, os.path.join(PROJECT_DIR, 'floor_texture'), floor)
        return floor_mesh


class ThreedFrontGroupDataset(Dataset):
    def __init__(self, room_type='bedroom', split='train', mode='order', batch_size=16, dataset_len=1000):
        # folders
        self.data_root = os.path.join(pathmagic.PROJECT_DIR, 'dataset')
        # self.room_type_dict = {'all': 'All', 'bedroom': 'Bedroom', 'diningroom': 'Diningroom', 'library': 'Library', 'livingroom': 'Livingroom'}
        self.data_folder = os.path.join(self.data_root, '3D-FRONT-labeled-group', split, room_type)
        self.group_data = {
            "Sleeping": [],
            "Relaxing": [],
            "BedWatchingTV": [],
            "Lighting": [],
            "Dressing": [],
            "ClothStorage": [],
            "BookStorage": [],
            "ShelfStorage": [],
            "KitchenStorage": [],
            "WatchingTV": [],
            "Working": [],
            "Eating": []
        }

        self.keys = ["Sleeping", "Relaxing", "BedWatchingTV", "Lighting", "Dressing", "ClothStorage", "BookStorage",
                     "ShelfStorage", "KitchenStorage", "WatchingTV", "Working", "Eating"]
        self.group_data = self.get_all_npy_file(self.group_data)
        self.good_index = []
        for i in range(len(self.keys)):
            if len(self.group_data[self.keys[i]]) != 0:
                self.good_index.append(i)
        self.good_index = np.array(self.good_index).astype('int16')

        self.mode = mode
        self.batch_size = batch_size
        self.dataset_len = dataset_len
        self.order_dict = {
            'bedroom': np.array([2, 0, 1, 10, 4, 8, 5, 3, 12, 15, 9, 11, 6, 7, 13, 14, 17, 16, 19, 20, 18, 22, 23, 21, 24, 25, 26]),
            'diningroom': np.array([8, 12, 10, 3, 5, 19, 14, 11, 6, 4, 18, 13, 25, 7, 9, 1, 2, 16, 22, 15, 0, 23, 20, 17, 21, 24, 26]),
            'library': np.array([6, 8, 9, 10, 4, 12, 13, 3, 14, 7, 11, 19, 0, 18, 1, 15, 20, 2, 22, 16, 5, 25, 23, 17, 21, 24, 26]),
            'livingroom': np.array([12, 8, 10, 5, 19, 3, 14, 11, 18, 6, 13, 4, 25, 7, 2, 1, 0, 9, 20, 16, 23, 22, 15, 17, 24, 21, 26]),
            'all': np.array([8, 12, 10, 2, 3, 0, 5, 6, 1, 19, 4, 14, 9, 11, 13, 18, 25, 7, 15, 16, 17, 20, 22, 23, 24, 21, 26])
        }
        self.order_list = self.order_dict[room_type]

    def __len__(self):
        return self.dataset_len

    def __getitem__(self, idx):
        np.random.seed()
        group_idx = np.random.choice(self.good_index, 1)[0]
        data_pool = self.group_data[self.keys[group_idx]]
        batch_idx = np.random.randint(0, len(data_pool), self.batch_size).astype('int16')
        bbox = []
        label = []
        bidx = []
        nins = []
        for i in range(self.batch_size):
            group_bbox = np.load(data_pool[batch_idx[i]])
            group_bbox[:, 8] = i
            group_bbox = self.order(group_bbox)
            bbox.append(group_bbox[:, 0:7])
            label.append(group_bbox[:, 7])
            bidx.append(group_bbox[:, 8])
            nins.append(group_bbox.shape[0])
        bbox = np.vstack(bbox)
        label = np.hstack(label)
        bidx = np.hstack(bidx)
        nins = np.hstack(nins)
        return bbox, label, bidx, nins, group_idx

    def get_group(self, group_idx):
        np.random.seed()
        data_pool = self.group_data[self.keys[group_idx]]
        data_idx = np.random.randint(0, len(data_pool))
        group_bbox = np.load(data_pool[data_idx])
        return group_bbox

    def order(self, group_bbox):
        label = group_bbox[:, 7].astype('int16')
        idx = np.argsort(np.argsort(self.order_list)[label])
        group_bbox = group_bbox[idx]
        # print(group_bbox[:, 7])
        return group_bbox

    def get_all_npy_file(self, group_data):
        filenames = os.listdir(self.data_folder)
        for filename in filenames:
            group_type = filename.split('_')[0]
            group_data[group_type].append(os.path.join(self.data_folder, filename))
        return group_data


if __name__ == '__main__':
    # test of scene dataset
    dataset = ThreedFrontDataset(room_type='bedroom', split='train', rot_aug=True)
    dataset.visualize_scene(100, show_mesh=True, return_vis=False)
    # test of semantic group dataset
    dataset = ThreedFrontGroupDataset(room_type='livingroom', mode='order', batch_size=3, dataset_len=1000)
    dataset.__getitem__(10)
