import pathmagic  # noqa
from scipy.spatial.transform import Rotation
from data.variables import CATEGORIES, SEMANTICS, GROUPS
import numpy as np
# import cv2 as cv
import vedo
import math
import json
import os


class UnityRoom:
    def __init__(self, json_file, room_mask_file, future_folder, floor_folder):
        self.json_file = json_file
        self.room_mask_file = room_mask_file
        # model folders
        self.future_folder = future_folder
        self.floor_folder = floor_folder
        # parse json info
        self.parse_json()
        # random color template
        np.random.seed(2022)
        self.color_temp = np.random.rand(10, 3)

    def parse_json(self):
        ''' json content:
                house_uid, room_uid, room_type, floor_obj, floor_center
                instances: instance_uid, instance_jid, category, object_index, postition, rotation, init_scale, scale
                instance_groups: instances_index
                label_groups
        '''
        with open(self.json_file, 'r') as f:
            room = json.load(f)
        f.close()
        self.house_uid = room['house_uid']
        self.room_uid = room['room_uid']
        self.room_type = room['room_type']
        self.floor_obj = room['floor_obj']
        self.floor_center = unity_vec3_2_array(room['floor_center'])
        # instances
        self.instance_uids = []
        self.instance_jids = []
        self.categories = []
        self.object_indexes = []
        self.positions = []
        self.rotations = []
        self.init_scales = []
        self.scales = []
        self.parse_instances(room['instances'])
        # group labels
        self.instance_groups = []
        self.parse_groups(room['instance_groups'])
        self.label_groups = room['label_groups']

    def parse_numpy_input_gt(self):
        ''' return:
                pos, rot, scale, category, object_index, groups
        '''
        # instances transform
        self.data_positions = np.array(self.positions)
        self.data_positions[:, 1] += np.array(self.init_scales)[:, 1] * 0.5
        self.data_rotations = np.array(self.rotations)
        self.data_scales = np.array(self.scales) * np.array(self.init_scales)
        self.data_scales[:, 0], self.data_scales[:, 2] = self.data_scales[:, 0] * 2, self.data_scales[:, 2] * 2
        # category
        categories = []
        for i in range(len(self.categories)):
            categories.append(SEMANTICS[CATEGORIES[self.categories[i]]])
        categories = np.array(categories)
        self.object_indexes = np.array(self.object_indexes)
        # groups
        group_labels = []
        for i in range(len(self.label_groups)):
            group_labels.append(GROUPS[self.label_groups[i]])
        group_labels = np.array(group_labels)
        return self.data_positions, self.data_rotations, self.data_scales, categories, self.object_indexes, self.instance_groups, group_labels

    def parse_instances(self, room_instances):
        for i in range(len(room_instances)):
            self.instance_uids.append(room_instances[i]['instance_uid'])
            self.instance_jids.append(room_instances[i]['instance_jid'])
            self.categories.append(room_instances[i]['category'])
            self.object_indexes.append(int(room_instances[i]['object_index']))
            self.positions.append(unity_vec3_2_array(room_instances[i]['position']))
            self.rotations.append(float(room_instances[i]['rotation']))
            self.init_scales.append(unity_vec3_2_array(room_instances[i]['init_scale']))
            self.scales.append(unity_vec3_2_array(room_instances[i]['scale']))

    def parse_groups(self, room_groups):
        for i in range(len(room_groups)):
            self.instance_groups.append(room_groups[i]['instances_index'])

    def unity_transform_mesh(self, mesh, pos, rot, scale):
        pts = mesh.points()
        pts = pts.astype(np.float64) * scale
        rotMatrix = Rotation.from_rotvec((rot / 360.) * 2 * math.pi * np.array([0, 1, 0])).as_matrix()
        R = np.array(rotMatrix)
        pts = np.transpose(pts)
        pts = np.matmul(R, pts)
        pts = np.transpose(pts)
        pts = pts + pos
        mesh = mesh.points(pts)

        arrow = np.matmul(R, np.array([0, 0, 1]).transpose())
        return mesh, arrow

    def visualize_room(self):
        vp = vedo.Plotter()
        self.vedo_mesh = []
        self.vedo_point = []
        self.vedo_arrow = []
        # floor
        floor_file = os.path.join(self.floor_folder, self.floor_obj.split('\\')[-1])
        floor_mesh = vedo.Mesh(floor_file)
        floor_mesh.color(np.array([1, 1, 1]))
        floor_mesh.points(floor_mesh.points() - self.floor_center)
        # instances
        for i in range(len(self.instance_uids)):
            mesh_file = os.path.join(self.future_folder, self.instance_jids[i], 'raw_model.obj')
            instance_mesh, arrow = self.unity_transform_mesh(vedo.Mesh(mesh_file), self.positions[i], self.rotations[i], self.scales[i])
            instance_mesh.color(np.array([1, 1, 1]))
            instance_mesh.alpha(0.5)
            instance_center_point = np.array(instance_mesh.points()).mean(0)
            instance_point = vedo.Point(pos=instance_center_point, r=20)
            instance_point.labels([f'{self.categories[i]}'])
            instance_arrow = vedo.Arrow(instance_center_point, instance_center_point+arrow)
            self.vedo_mesh.append(instance_mesh)
            self.vedo_point.append(instance_point)
            self.vedo_arrow.append(instance_arrow)
        # add slider
        self.vedo_mesh.append(floor_mesh)
        vp.addSlider2D(self.vedo_slider, 0, len(self.label_groups)-1, value=0, pos='bottom-right', title='semantic groups')
        vp.show(self.vedo_mesh + self.vedo_point + self.vedo_arrow, axes=2, interactive=1)

    def vedo_slider(self, widget, event):
        value = widget.GetRepresentation().GetValue()
        curr_group = self.instance_groups[int(value)]
        for i in range(len(self.instance_uids)):
            if self.object_indexes[i] in curr_group:
                self.vedo_mesh[i].color(np.array([0, 1, 0]))
            else:
                self.vedo_mesh[i].color(np.array([1, 1, 1]))
        print(self.label_groups[int(value)])


def unity_vec3_2_array(unity_vector3):
    return np.array([float(unity_vector3['x']), float(unity_vector3['y']), float(unity_vector3['z'])])
