from data.variables import GROUP_COLOR, GROUPS, SEMANTICS, SEMANTICS_COLOR, FUTURE_FOLDER
from scipy.spatial.transform import Rotation
import numpy as np
import random
import vedo
import math
import cv2
import os


def bbox_to_lines(bbox: vedo.Points, lw=5, c='k4'):
    start_idx = [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3]
    end_idx = [1, 2, 3, 0, 5, 6, 7, 4, 4, 5, 6, 7]
    return vedo.Lines(bbox.points()[start_idx], bbox.points()[end_idx], lw=lw, c=c)


def unit_bbox_generator(r=10):
    return vedo.Points(
        np.array([
            [0.5, -0.5, 0.5],
            [0.5, -0.5, -0.5],
            [-0.5, -0.5, -0.5],
            [-0.5, -0.5, 0.5],
            [0.5, 0.5, 0.5],
            [0.5, 0.5, -0.5],
            [-0.5, 0.5, -0.5],
            [-0.5, 0.5, 0.5]
        ]),
        r=r
    )


def transform_points(unit_bbox: vedo.Points, pos, rot, scale):
    pts = unit_bbox.points()
    if scale is not None:
        pts = pts.astype(np.float64) * scale
    rotMatrix = Rotation.from_rotvec((rot / 360.) * 2 * math.pi * np.array([0, 1, 0])).as_matrix()
    R = np.array(rotMatrix)
    pts = np.transpose(pts)
    pts = np.matmul(R, pts)
    pts = np.transpose(pts)
    pts = pts + pos
    unit_bbox = unit_bbox.points(pts)

    arrow = np.matmul(R, np.array([0, 0, 1]).transpose())
    return unit_bbox, arrow


def unity_transform_mesh(mesh, pos, rot, scale):
    pts = mesh.points().astype(np.float64)
    pts[:, 1] = pts[:, 1] - (pts.max(0) + pts.min(0))[1] / 2
    relative_scale = scale / (pts.max(0) - pts.min(0))
    pts = pts * relative_scale
    rotMatrix = Rotation.from_rotvec((rot / 360.) * 2 * math.pi * np.array([0, 1, 0])).as_matrix()
    R = np.array(rotMatrix)
    pts = np.transpose(pts)
    pts = np.matmul(R, pts)
    pts = np.transpose(pts)
    pts = pts + pos
    mesh = mesh.points(pts)
    return mesh


def load_floor(room, tex_folder=None, idx=None):
    floor_file = os.path.join(room.floor_folder, room.floor_obj.split('\\')[-1])
    floor_mesh = vedo.Mesh(floor_file)
    if tex_folder is not None:
        file_list = os.listdir(tex_folder)
        if idx is None:
            np.random.seed()
            idx = np.random.randint(0, len(file_list))
        floor_mesh.texture(os.path.join(tex_folder, file_list[idx]), repeat=True, scale=3)
    floor_mesh.color(np.array([1, 1, 1]))
    floor_mesh.points(floor_mesh.points() - room.floor_center)
    return floor_mesh


def load_floor_fastsynth(objfile, tex_folder=None, idx=None):
    floor_mesh = vedo.Mesh(objfile)
    floor_center = floor_mesh.points().mean(0)
    if tex_folder is not None:
        file_list = os.listdir(tex_folder)
        if idx is None:
            np.random.seed()
            idx = np.random.randint(0, len(file_list))
        floor_mesh.texture(os.path.join(tex_folder, file_list[idx]), repeat=True, scale=3)
    floor_mesh.color(np.array([1, 1, 1]))
    floor_mesh.points(floor_mesh.points() - floor_center)
    return floor_mesh


def visualize_scene(group_bbox, group_lab, normalized=True, mode='proposal', frame=True, numpy=False, only_model=True, model_pool=None, c='k4', rotate=False):
    if numpy:
        bbox = group_bbox
        label = group_lab.astype('int16')
    else:
        bbox = group_bbox.detach().cpu().numpy()
        label = group_lab.detach().cpu().numpy().astype('int16')
    if normalized:
        pos = bbox[:, 0:3] * 4
        rot = bbox[:, 3] * 180 + 180
        if mode == 'complete':
            scale = bbox[:, 4:7] * 4 + 4
    else:
        pos = bbox[:, :3]
        rot = bbox[:, 3]
        if mode == 'complete':
            scale = bbox[:, 4:7]
    vedo_vis, vedo_vis_bbox_frame, vis_model = [], [], []
    if frame:
        frame_bbox = unit_bbox_generator(r=10)
        frame_bbox, _ = transform_points(frame_bbox, np.array([0, 2, 0]), np.zeros(1), np.array([8, 4, 8]))
        frame_line = bbox_to_lines(frame_bbox, lw=5, c=c)
        vedo_vis.append(frame_bbox)
        vedo_vis.append(frame_line)
    if model_pool is not None:
        model_dict = {}
        for key in model_pool.keys():
            model_dict[key] = random.choice(model_pool[key])
    for i in range(len(pos)):
        unit_bbox = unit_bbox_generator(r=10)
        if mode == 'complete':
            bbox, arrow = transform_points(unit_bbox, pos[i], rot[i], scale[i])
            color = SEMANTICS_COLOR[label[i]]
            lines = bbox_to_lines(bbox, lw=5, c=c)
            arrow = vedo.Arrow(pos[i], pos[i]+arrow)
            center = vedo.Sphere(pos[i], r=0.1)
            arrow.color(color)
            center.color(color)
            if not only_model:
                vedo_vis_bbox_frame.append(bbox)
                vedo_vis_bbox_frame.append(lines)
                vedo_vis_bbox_frame.append(arrow)
            if model_pool is not None:
                mesh_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'raw_model.obj')
                tex_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'texture.png')
                mesh_obj = vedo.Mesh(mesh_file)
                tex = cv2.imread(tex_file)
                mesh_obj.texture(tex)
                instance_mesh = unity_transform_mesh(mesh_obj, pos[i], rot[i], scale[i])
                instance_mesh.alpha(1)
                vedo_vis.append(instance_mesh)
                if not vis_model:
                    vis_model = instance_mesh
                else:
                    vis_model += instance_mesh
        elif mode == 'proposal':
            bbox, arrow = transform_points(unit_bbox, pos[i], rot[i], None)
            color = GROUP_COLOR[label[i]]
            arrow = vedo.Arrow(pos[i], pos[i]+arrow)
            center = vedo.Sphere(pos[i], r=0.1)
            arrow.color(color)
            center.color(color)
            vedo_vis.append(center)
            vedo_vis.append(arrow)
    string_label = []
    for i in range(label.shape[0]):
        if mode == 'complete':
            string_label.append(list(SEMANTICS.keys())[label[i]])
        elif mode == 'proposal':
            string_label.append(list(GROUPS.keys())[label[i]])
    pts = vedo.Points(pos, r=5)
    vedo_lab = pts.labels(string_label)
    if only_model:
        if rotate:
            return vis_model
        else:
            return vedo_vis, vedo_lab
    else:
        vedo_vis_bbox_frame.append(pts)
        if rotate:
            return vis_model
        else:
            return vedo_vis, vedo_vis_bbox_frame, vedo_lab


def visualize_proposenet_scene(group_bbox, normalized=True, color='k4'):
    bbox = group_bbox.detach().cpu().numpy()
    if normalized:
        pos = bbox[:, 0:3] * 4
        rot = bbox[:, 3] * 180 + 180
    else:
        pos = bbox[:, :3]
        rot = bbox[:, 3]
    vedo_vis = []
    frame_bbox = unit_bbox_generator(r=10)
    frame_bbox, _ = transform_points(frame_bbox, np.array([0, 2, 0]), np.zeros(1), np.array([8, 4, 8]))
    frame_line = bbox_to_lines(frame_bbox, lw=5)
    vedo_vis.append(frame_bbox)
    vedo_vis.append(frame_line)
    for i in range(len(pos)):
        unit_bbox = unit_bbox_generator(r=10)
        bbox, arrow = transform_points(unit_bbox, pos[i], rot[i], None)
        arrow = vedo.Arrow(pos[i], pos[i]+arrow)
        center = vedo.Sphere(pos[i], r=0.15)
        arrow.color(color)
        center.color(color)
        vedo_vis.append(center)
        vedo_vis.append(arrow)
    return vedo_vis


def visualize_ours_output(group_bbox, group_bbox_wo_opt, group_lab, model_pool):
    bbox = group_bbox
    bbox_wo_opt = group_bbox_wo_opt
    label = group_lab.astype('int16')
    pos = bbox[:, 0:3] * 4
    rot = bbox[:, 3] * 180 + 180
    scale = bbox[:, 4:7] * 4 + 4
    pos_wo_opt = bbox_wo_opt[:, 0:3] * 4
    rot_wo_opt = bbox_wo_opt[:, 3] * 180 + 180
    scale_wo_opt = bbox_wo_opt[:, 4:7] * 4 + 4

    vedo_vis, vedo_vis_wo_opt = [], []
    if model_pool is not None:
        model_dict = {}
        for key in model_pool.keys():
            model_dict[key] = random.choice(model_pool[key])
    for i in range(len(pos)):
        mesh_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'raw_model.obj')
        tex_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'texture.png')
        mesh_obj_opt = vedo.Mesh(mesh_file)
        tex_opt = cv2.imread(tex_file)
        mesh_obj_opt.texture(tex_opt)
        mesh_obj_wo_opt = vedo.Mesh(mesh_file)
        tex_wo_opt = cv2.imread(tex_file)
        mesh_obj_wo_opt.texture(tex_wo_opt)
        instance_mesh_opt = unity_transform_mesh(mesh_obj_opt, pos[i], rot[i], scale[i])
        instance_mesh_opt.alpha(1)
        instance_mesh_wo_opt = unity_transform_mesh(mesh_obj_wo_opt, pos_wo_opt[i], rot_wo_opt[i], scale_wo_opt[i])
        instance_mesh_wo_opt.alpha(1)
        vedo_vis.append(instance_mesh_opt)
        vedo_vis_wo_opt.append(instance_mesh_wo_opt)
    return vedo_vis, vedo_vis_wo_opt


def visualize_synthetic_scene(vp, group_bbox, group_lab, normalized=True, model_pool=None, numpy=False, vis=False, c='k4'):
    if numpy:
        bbox = group_bbox
        label = group_lab.astype('int16')
    else:
        bbox = group_bbox.detach().cpu().numpy()
        label = group_lab.detach().cpu().numpy().astype('int16')
    if normalized:
        pos = bbox[:, 0:3] * 4
        rot = bbox[:, 3] * 180 + 180
        scale = bbox[:, 4:7] * 4 + 4
    else:
        pos = bbox[:, :3]
        rot = bbox[:, 3]
        scale = bbox[:, 4:7]
    if model_pool is not None:
        model_dict = {}
        for key in model_pool.keys():
            model_dict[key] = random.choice(model_pool[key])
    vis_object = []
    for i in range(len(pos)):
        mesh_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'raw_model.obj')
        tex_file = os.path.join(FUTURE_FOLDER, model_dict[str(label[i])], 'texture.png')
        mesh_obj = vedo.Mesh(mesh_file)
        tex = cv2.imread(tex_file)
        mesh_obj.texture(tex)
        instance_mesh = unity_transform_mesh(mesh_obj, pos[i], rot[i], scale[i])
        instance_mesh.alpha(1)
        if not vis:
            vp += instance_mesh
        if not vis_object:
            vis_object = instance_mesh
        else:
            vis_object += instance_mesh
    if not vis:
        return vp
    else:
        return vis_object
