from utils import MLP, ConditionalMAF, ConditionalNormalizingFlowModel
from utils import SemanticGraph, GraphAttnEncoder
from torch.distributions import MultivariateNormal
import torch.nn as nn
import torch


class CompleteNet(nn.Module):
    def __init__(self, ndim=128, ndimout=128, nclass=27, nlayer=3, mode='ae'):
        super(CompleteNet, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.ndimout = ndimout
        self.ndimcon = ndim

        self.mode = mode

        self.complete_encoder = CompleteNet_GraphEncoder(ndim=self.ndim, nclass=self.nclass, nlayer=self.nlayer)
        self.complete_decoder = CompleteNet_GraphAllDecoder(ndim=self.ndim, nclass=self.nclass, nlayer=self.nlayer)
        self.global_feat_generator = ConditionalFeatureGenerator(ndim=self.ndim, ndimcon=10, nclass=self.nclass, nlayer=8)

    def forward(self, group_bbox, group_onehot, N):
        node_feat = self.complete_encoder(group_bbox, group_onehot, N)
        if self.mode == 'ae':
            global_feat = torch.sum(node_feat, dim=0, keepdim=True)
            node_feat = self.complete_decoder(global_feat, N)
            return node_feat
        elif self.mode == 'cmaf':
            global_feat = torch.sum(node_feat, dim=0, keepdim=True)
            node_feat = self.complete_decoder(global_feat, N)
            condition = torch.eye(10)[torch.clamp(torch.Tensor([N-1]).long().cuda(), min=0, max=9)].cuda()
            logprob = self.global_feat_generator(global_feat, condition)
            return node_feat, logprob


class CompleteNet_GraphEncoder(nn.Module):
    def __init__(self, ndim=128, nclass=27, nlayer=3):
        super(CompleteNet_GraphEncoder, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.posrot_encoder = MLP(mlp=[7, self.ndim//4, self.ndim//2])
        self.onehot_encoder = MLP(mlp=[self.nclass, self.ndim//4, self.ndim//2])
        self.graph_encoder = GraphAttnEncoder(ndim=self.ndim, nlayer=self.nlayer)

    def forward(self, group_bbox, group_onehot, N):
        bbox_feat = self.posrot_encoder(group_bbox)
        onehot_feat = self.onehot_encoder(group_onehot)
        node_feat = torch.cat([bbox_feat, onehot_feat], dim=-1)
        if N == 1:
            return node_feat
        if N != 1:
            edges = self.generate_edges(N-1)
            G = SemanticGraph(node_feat, edges)
            node_feat = self.graph_encoder(G)
        return node_feat

    def generate_edges(self, idx):
        edges = []
        for i in range(0, idx+1):
            for j in range(0, idx+1):
                if i != j:
                    edges.append([i, j])
        edges = torch.Tensor(edges).long().cuda().transpose(0, 1).contiguous()
        return edges


class CompleteNet_GraphBatchEncoder(nn.Module):
    def __init__(self, ndim=128, nclass=27, nlayer=3):
        super(CompleteNet_GraphBatchEncoder, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.posrot_encoder = MLP(mlp=[7, self.ndim//4, self.ndim//2])
        self.onehot_encoder = MLP(mlp=[self.nclass, self.ndim//4, self.ndim//2])
        self.graph_encoder = GraphAttnEncoder(ndim=self.ndim, nlayer=self.nlayer)

    def forward(self, bbox, onehot, nins):
        bbox_feat = self.posrot_encoder(bbox)
        onehot_feat = self.onehot_encoder(onehot)
        node_feat = torch.cat([bbox_feat, onehot_feat], dim=-1)
        if nins.shape[0] == nins.sum():
            return node_feat
        else:
            edges = self.generate_edges(nins)
            G = SemanticGraph(node_feat, edges)
            node_feat = self.graph_encoder(G)
        return node_feat

    def generate_edges(self, nins):
        N = nins.sum()
        edges = torch.zeros(N, N).cuda()
        sum = 0
        for i in range(nins.shape[0]):
            edges[sum:sum+nins[i], sum:sum+nins[i]] = 1
            sum = sum + nins[i]
        edges = edges - torch.eye(N).cuda()
        edges = torch.nonzero(edges).transpose(0, 1)
        return edges


class CompleteNet_GraphAllDecoder(nn.Module):
    def __init__(self, ndim=128, nclass=27, nlayer=3):
        super(CompleteNet_GraphAllDecoder, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.graph_decoder = GraphAttnEncoder(ndim=self.ndim, nlayer=self.nlayer)
        # self.new_node_decoder = MLP(mlp=[self.ndim, self.ndim, self.ndim])
        self.new_node_decoder = MLP(mlp=[self.ndim * 2, self.ndim, self.ndim])

    def forward(self, global_feat, N):
        for i in range(N):
            if i == 0:
                # node_feat = self.new_node_decoder(global_feat)
                node_feat = self.new_node_decoder(torch.cat([global_feat, global_feat], dim=-1))
            else:
                # new_node = self.new_node_decoder(node_feat[-1, :].view(1, -1).contiguous())
                new_node = self.new_node_decoder(torch.cat([global_feat, node_feat[-1, :].view(1, -1).contiguous()], dim=-1))
                node_feat = torch.cat([node_feat, new_node], dim=0)
                edges = self.generate_edges(node_feat.shape[0]-1)
                G = SemanticGraph(node_feat, edges)
                node_feat = self.graph_decoder(G)
        return node_feat

    def generate_edges(self, idx):
        edges = []
        for i in range(0, idx+1):
            for j in range(0, idx+1):
                if i != j:
                    edges.append([i, j])
        edges = torch.Tensor(edges).long().cuda().transpose(0, 1).contiguous()
        return edges


class CompleteNet_GraphOrderDecoder(nn.Module):
    def __init__(self, ndim=128, nclass=27, nlayer=3):
        super(CompleteNet_GraphOrderDecoder, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.graph_decoder = GraphAttnEncoder(ndim=self.ndim+10, nlayer=self.nlayer)
        self.node_decoder = MLP(mlp=[self.ndim+10, self.ndim, self.ndim])

    def forward(self, global_feat, N):
        node_feat = global_feat.repeat(N, 1).view(N, -1).contiguous()
        idx_feat = torch.eye(10)[torch.arange(N)].cuda()
        node_feat = torch.cat([node_feat, idx_feat], dim=-1)
        if node_feat.shape[0] == 1:
            node_feat = self.node_decoder(node_feat)
            return node_feat
        else:
            edges = self.generate_edges(node_feat.shape[0]-1)
            G = SemanticGraph(node_feat, edges)
            node_feat = self.graph_decoder(G)
            node_feat = self.node_decoder(node_feat)
            return node_feat

    def generate_edges(self, idx):
        edges = []
        for i in range(0, idx+1):
            for j in range(0, idx+1):
                if i != j:
                    edges.append([i, j])
        edges = torch.Tensor(edges).long().cuda().transpose(0, 1).contiguous()
        return edges


class CompleteNet_GraphDecoder(nn.Module):
    def __init__(self, ndim=128, nclass=27, nlayer=3):
        super(CompleteNet_GraphDecoder, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.nlayer = nlayer
        self.graph_decoder = GraphAttnEncoder(ndim=self.ndim+self.ndim, nlayer=self.nlayer)
        self.node_decoder = MLP(mlp=[self.ndim+self.ndim, self.ndim, self.ndim])

    def forward(self, global_feat, node_feat, N):
        feat = global_feat.repeat(N, 1).view(N, -1).contiguous()
        feat = torch.cat([feat, node_feat], dim=-1)
        if feat.shape[0] == 1:
            node_feat = self.node_decoder(feat)
            return node_feat
        else:
            edges = self.generate_edges(feat.shape[0]-1)
            G = SemanticGraph(feat, edges)
            node_feat = self.graph_decoder(G)
            node_feat = self.node_decoder(node_feat)
            return node_feat

    def generate_edges(self, idx):
        edges = []
        for i in range(0, idx+1):
            for j in range(0, idx+1):
                if i != j:
                    edges.append([i, j])
        edges = torch.Tensor(edges).long().cuda().transpose(0, 1).contiguous()
        return edges


class VAE_Parameters_Encoder(nn.Module):
    def __init__(self, ndim=64):
        super(VAE_Parameters_Encoder, self).__init__()
        self.ndim = ndim
        self.mu_encoder = MLP(mlp=[self.ndim, self.ndim, self.ndim])
        self.logstd_encoder = MLP(mlp=[self.ndim, self.ndim, self.ndim])

    def forward(self, graph_feat):
        mu = self.mu_encoder(graph_feat)
        logstd = self.logstd_encoder(graph_feat)
        return mu, logstd


class CategoryDiscretDecoder(nn.Module):
    def __init__(self, ndim=64, ndimout=128, nclass=27):
        super(CategoryDiscretDecoder, self).__init__()
        self.ndim = ndim
        self.ndimout = ndimout
        self.nclass = nclass

        # predict attributes fc layers
        self.fc_pred_class = MLP(mlp=[self.ndim, self.ndim, self.nclass], output='log_sm')

    def forward(self, node_feat):
        output = {}
        output['class'] = self.fc_pred_class(node_feat)
        return output


class AttributeDiscretDecoder(nn.Module):
    def __init__(self, ndim=64, ndimout=128, nclass=27):
        super(AttributeDiscretDecoder, self).__init__()
        self.ndim = ndim
        self.ndimout = ndimout
        self.nclass = nclass

        # predict attributes fc layers
        self.fc_pred_class = MLP(mlp=[self.ndim, self.ndim, self.nclass], output='log_sm')

        hidden_size = self.nclass + self.ndim
        self.fc_pred_pos_x = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')
        self.fc_pred_pos_y = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')
        self.fc_pred_pos_z = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')

        self.fc_pred_rot = MLP(mlp=[hidden_size, self.ndim, 8], output='log_sm')

        self.fc_pred_scale_x = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')
        self.fc_pred_scale_y = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')
        self.fc_pred_scale_z = MLP(mlp=[hidden_size, self.ndim, self.ndimout], output='log_sm')

    def forward(self, node_feat):
        output = {}
        output['class'] = self.fc_pred_class(node_feat)
        node_feat = torch.cat([output['class'], node_feat], dim=-1)
        output['px'] = self.fc_pred_pos_x(node_feat)
        output['py'] = self.fc_pred_pos_y(node_feat)
        output['pz'] = self.fc_pred_pos_z(node_feat)
        output['r'] = self.fc_pred_rot(node_feat)
        output['sx'] = self.fc_pred_scale_x(node_feat)
        output['sy'] = self.fc_pred_scale_y(node_feat)
        output['sz'] = self.fc_pred_scale_z(node_feat)
        return output


class ConditionalFeatureGenerator(nn.Module):
    def __init__(self, ndim, ndimcon, nclass, nlayer):
        super(ConditionalFeatureGenerator, self).__init__()
        self.ndim = ndim
        self.ndimcon = ndimcon
        self.nclass = nclass
        self.nlayer = nlayer

        self.prior = MultivariateNormal(torch.zeros(self.ndim).cuda(), torch.eye(self.ndim).cuda())
        self.flows = [ConditionalMAF(self.ndim, condition_dim=self.ndimcon) for _ in range(self.nlayer)]
        self.cnf = ConditionalNormalizingFlowModel(self.prior, self.flows).cuda()

    def forward(self, feat, condition):
        zs, prior_logprob, log_det = self.cnf(feat, condition)
        logprob = prior_logprob + log_det
        return logprob

    def sample(self, condition):
        N = condition.shape[0]
        feat = self.cnf.sample(N, condition)
        return feat


class ProposeNetAutoencoder(nn.Module):
    def __init__(self, ndim, ndimcon, ndimout, nclass, nlayer):
        super(ProposeNetAutoencoder, self).__init__()
        self.ndim = ndim
        self.ndimcon = ndimcon
        self.ndimout = ndimout
        self.nclass = nclass
        self.nlayer = nlayer

        self.posrot_encoder = MLP(mlp=[4, self.ndim//2, self.ndim//2])
        self.onehot_encoder = MLP(mlp=[nclass, self.ndim//2, self.ndim//2])
        self.decoder = AttributeDiscretDecoder(self.ndim, self.ndimout, self.nclass, mode='propose_net')

        self.prior = MultivariateNormal(torch.zeros(self.ndim).cuda(), torch.eye(self.ndim).cuda())
        self.flows = [ConditionalMAF(self.ndim, condition_dim=self.ndimcon) for _ in range(self.nlayer)]
        self.cnf = ConditionalNormalizingFlowModel(self.prior, self.flows).cuda()

    def forward(self, posrot, onehot):
        feat_posrot = self.posrot_encoder(posrot)
        feat_onehot = self.onehot_encoder(onehot)
        feat = torch.cat([feat_onehot, feat_posrot], dim=-1)
        output = self.decoder(feat)
        zs, prior_logprob, log_det = self.cnf(feat, onehot)
        logprob = prior_logprob + log_det
        return output, logprob

    def sample(self, condition):
        N = condition.shape[0]
        feat = self.cnf.sample(N, condition)
        output = self.decoder(feat)
        return output


class ProposeNet(nn.Module):
    def __init__(self, ndim, ndimcon, nlayer):
        super(ProposeNet, self).__init__()
        self.ndim = ndim
        self.ndimcon = ndimcon
        self.nlayer = nlayer

        self.prior = MultivariateNormal(torch.zeros(self.ndim).cuda(), torch.eye(self.ndim).cuda())
        self.flows = [ConditionalMAF(self.ndim, condition_dim=self.ndimcon) for _ in range(self.nlayer)]
        self.cnf = ConditionalNormalizingFlowModel(self.prior, self.flows).cuda()

    def forward(self, feat, condition):
        zs, prior_logprob, log_det = self.cnf(feat, condition)
        logprob = prior_logprob + log_det
        return logprob

    def sample(self, condition):
        N = condition.shape[0]
        feat = self.cnf.sample(N, condition)
        return feat


if __name__ == '__main__':
    pass
