import pathmagic  # noqa
from data.data_loader import ThreedFrontGroupDataset
import data.visualizer as visualizer  # noqa
from scipy.spatial.transform import Rotation
from torch.utils.data import DataLoader
from model import CompleteNet_GraphBatchEncoder, VAE_Parameters_Encoder, CompleteNet_GraphAllDecoder, AttributeDiscretDecoder
import torch.nn.functional as F  # noqa
import torch.nn as nn
import numpy as np
import torch
import math
import time
import vedo  # noqa
import os


os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'
MAX_LOGSTD = 10


class get_model(nn.Module):
    def __init__(self, ndim=128, ndimout=64, nclass=27):
        super(get_model, self).__init__()
        self.ndim = ndim
        self.ndimout = ndimout
        self.nclass = nclass
        self.complete_encoder = CompleteNet_GraphBatchEncoder(ndim=self.ndim, nclass=self.nclass, nlayer=3)
        self.vae_encoder = VAE_Parameters_Encoder(self.ndim)
        self.complete_decoder = CompleteNet_GraphAllDecoder(ndim=self.ndim+12+10, nclass=self.nclass, nlayer=3)
        self.output = AttributeDiscretDecoder(ndim=self.ndim+12+10, ndimout=self.ndimout, nclass=self.nclass)

    def forward(self, bbox, label, bidx, nins, group_idx):
        B = bidx.max() + 1
        onehot = torch.eye(self.nclass)[label].cuda()
        node_feat = self.complete_encoder(bbox, onehot, nins)
        # node_feat -> graph_feat
        b_weight = torch.eye(B)[bidx].cuda()
        graph_feat = torch.matmul(node_feat.transpose(0, 1), b_weight).transpose(0, 1)
        # vae encoding
        mu, logstd = self.vae_encoder(graph_feat)
        n_condition = torch.eye(10)[torch.clamp(nins, min=0, max=9)].cuda()
        g_condition = torch.eye(12)[group_idx].repeat(B, 1).cuda()
        global_feat = self.reparametrize(mu, logstd)
        global_feat = torch.cat([global_feat, g_condition, n_condition], dim=-1)
        # graph_feat -> node_feat
        dec_feat = []
        for i in range(B):
            node_dec_feat = self.complete_decoder(global_feat[i].view(1, -1).contiguous(), nins[i])
            dec_feat.append(node_dec_feat)
        dec_feat = torch.cat(dec_feat, dim=0)
        pred_output = self.output(dec_feat)
        gt_output = self.generate_gt(bbox, label)
        return pred_output, gt_output, mu, logstd

    def reparametrize(self, mu, logstd):
        if self.training:
            return mu + torch.randn_like(logstd) * torch.exp(logstd)

    def sample(self, group_idx, nins):
        n_condition = torch.eye(10)[torch.clamp(nins, min=0, max=9)].cuda().view(1, -1).contiguous()
        g_condition = torch.eye(12)[group_idx].cuda().view(1, -1).contiguous()
        global_feat = torch.randn(1, self.ndim).cuda()
        global_feat = torch.cat([global_feat, g_condition, n_condition], dim=-1)
        node_dec_feat = self.complete_decoder(global_feat, nins)
        pred_output = self.output(node_dec_feat)
        return pred_output

    def sample_from_data(self, bbox, label, group_idx, nins):
        condition = torch.eye(10)[torch.clamp(nins, min=0, max=10)].cuda()
        graph_feat = self.cmaf[group_idx].sample(condition)
        node_dec_feat = self.complete_decoder[group_idx](graph_feat.view(1, -1).contiguous(), nins)
        pred_output = self.output(node_dec_feat)
        gt_output = self.generate_gt(bbox, label)
        return pred_output, gt_output

    def generate_gt(self, group_posrot, group_lab):
        output = {}
        # class
        output['class'] = group_lab
        # position
        offset = (self.ndimout - 1) * 0.5
        output['px'] = torch.clamp(torch.round((group_posrot[:, 0] * offset) + offset).long(), min=0, max=self.ndimout-1)
        output['py'] = torch.clamp(torch.round((group_posrot[:, 1] * offset) + offset).long(), min=0, max=self.ndimout-1)
        output['pz'] = torch.clamp(torch.round((group_posrot[:, 2] * offset) + offset).long(), min=0, max=self.ndimout-1)
        # rotation
        rot = torch.clamp(torch.round((group_posrot[:, 3] * 4) + 4).long(), min=0, max=8)
        idx = rot == 8
        rot[idx] = 0
        output['r'] = rot
        # scale
        output['sx'] = torch.clamp(torch.round((group_posrot[:, 4] * offset) + offset).long(), min=0, max=self.ndimout-1)
        output['sy'] = torch.clamp(torch.round((group_posrot[:, 5] * offset) + offset).long(), min=0, max=self.ndimout-1)
        output['sz'] = torch.clamp(torch.round((group_posrot[:, 6] * offset) + offset).long(), min=0, max=self.ndimout-1)
        return output

    def continuous_output(self, output, type='pred'):
        keys = ['px', 'py', 'pz', 'r', 'sx', 'sy', 'sz']
        offset = (self.ndimout - 1) * 0.5
        bbox = []
        if type == 'pred':
            for i in range(len(keys)):
                if i == 3:
                    data = (output[keys[i]].topk(k=1).indices.view(-1, 1).contiguous() - 4) / 4
                else:
                    data = (output[keys[i]].topk(k=1).indices.view(-1, 1).contiguous() - offset) / offset
                bbox.append(data)
        elif type == 'gt':
            for i in range(len(keys)):
                if i == 3:
                    data = (output[keys[i]].view(-1, 1).contiguous() - 4) / 4
                else:
                    data = (output[keys[i]].view(-1, 1).contiguous() - offset) / offset
                bbox.append(data)
        bbox = torch.cat(bbox, dim=1)
        return bbox

    def shuffle_group(self, group_bbox, group_onehot):
        idx = np.arange(group_bbox.shape[0])
        np.random.seed()
        np.random.shuffle(idx)
        idx = torch.LongTensor(idx).cuda()
        return group_bbox[idx], group_onehot[idx]

    def order_group(self, group_bbox, group_onehot):
        idx = group_onehot.topk(k=1).indices.view(-1).argsort()
        return group_bbox[idx], group_onehot[idx]

    def normalize_group(self, group_bbox):
        # get the biggest instance and its rot
        core_inst_idx = (group_bbox[:, 4] * group_bbox[:, 5] * group_bbox[:, 6]).argmax()
        core_rot = group_bbox[core_inst_idx, 3].clone()
        group_bbox[:, 3] = group_bbox[:, 3] - group_bbox[core_inst_idx, 3]
        # get the offset
        pos = group_bbox[:, :3]
        maxs = torch.max(pos, dim=0).values
        mins = torch.min(pos, dim=0).values
        offset = (maxs + mins) * 0.5
        offset[1] = 0
        pos = pos - offset
        # get the rotation
        rotMatrix = Rotation.from_rotvec((core_rot.item() / 360.) * 2 * math.pi * np.array([0, -1, 0])).as_matrix()
        R = torch.Tensor(np.array(rotMatrix)).cuda()
        pos = torch.matmul(R, pos.transpose(0, 1)).transpose(0, 1)
        group_bbox[:, :3] = pos
        # normalize
        # pos [-4, 4]->[-1, 1]; rot [0, 360]->[-1, 1]; scale [0, 8]->[-1, 1]
        group_bbox[:, 0:3] = group_bbox[:, 0:3] / 4
        group_bbox[:, 3] = ((group_bbox[:, 3] % 360) - 180.0) / 180.0
        group_bbox[:, 4:7] = (group_bbox[:, 4:7] - 4) / 4
        return group_bbox


class get_loss(nn.Module):
    def __init__(self):
        super(get_loss, self).__init__()
        self.w = [1, 1, 0.3]

    def forward(self, pred_output, gt_output, mu, logstd):
        loss_class = F.nll_loss(pred_output['class'], gt_output['class'], reduction='mean')
        loss_bbox = self.get_bbox_loss(pred_output, gt_output)
        loss_vae = self.kl_loss(mu, logstd)
        loss = self.w[0]*loss_class + self.w[1]*loss_bbox + self.w[2]*loss_vae
        return loss, loss_class, loss_bbox, loss_vae

    def get_bbox_loss(self, pred_output, gt_output):
        loss_px = F.nll_loss(pred_output['px'], gt_output['px'], reduction='mean')
        loss_py = F.nll_loss(pred_output['py'], gt_output['py'], reduction='mean')
        loss_pz = F.nll_loss(pred_output['pz'], gt_output['pz'], reduction='mean')
        loss_r = F.nll_loss(pred_output['r'], gt_output['r'], reduction='mean')
        loss_sx = F.nll_loss(pred_output['sx'], gt_output['sx'], reduction='mean')
        loss_sy = F.nll_loss(pred_output['sy'], gt_output['sy'], reduction='mean')
        loss_sz = F.nll_loss(pred_output['sz'], gt_output['sz'], reduction='mean')
        loss_bbox = (loss_px + loss_py + loss_pz + loss_r + loss_sx + loss_sy + loss_sz)
        return loss_bbox

    def get_cmaf_loss(self, logprob):
        return -torch.mean(logprob)

    def kl_loss(self, mu=None, logstd=None):
        mu = self.__mu__ if mu is None else mu
        logstd = self.__logstd__ if logstd is None else logstd.clamp(
            max=MAX_LOGSTD)
        return -0.5 * torch.mean(
            torch.sum(1 + 2 * logstd - mu**2 - logstd.exp()**2, dim=1))


if __name__ == '__main__':
    from tqdm import tqdm
    TRAINING_SET = ThreedFrontGroupDataset(room_type='bedroom', mode='order', batch_size=3, dataset_len=1000)
    trainDataLoader = DataLoader(TRAINING_SET, batch_size=1, shuffle=False, num_workers=0, pin_memory=True,
                                 drop_last=True, worker_init_fn=lambda x: np.random.seed(x+int(time.time())))
    network = get_model().cuda()
    criterion = get_loss().cuda()
    bar = tqdm(enumerate(trainDataLoader), total=len(trainDataLoader), smoothing=0.9)
    for i, data in bar:
        bbox, label, bidx, nins, group_idx = data
        bbox = bbox[0].float().cuda()
        label = label[0].long().cuda()
        bidx = bidx[0].long().cuda()
        nins = nins[0].long().cuda()
        group_idx = group_idx[0].long().cuda()
        # print(bbox.shape, label, bidx, nins, group_idx)
        # all_bbox = all_bbox[0].float().cuda()
        # bbox_onehot = bbox_onehot[0].float().cuda()
        # group_lab = group_lab[0].long().cuda()

        # print(all_bbox.shape, bbox_onehot.shape, group_lab)
        pred_output, gt_output, mu, logstd = network(bbox, label, bidx, nins, group_idx)

        # # vis, vis_lab = network.visualize_scene(group_posrot[1:-1], group_lab[1:-1]-1, normalized=True)
        # # vedo.show(vis, vis_lab, axes=2, interactive=1)
        loss, loss_class, loss_bbox, loss_cmaf = criterion(pred_output, gt_output, mu, logstd)
        print(loss, loss_class, loss_bbox, loss_cmaf)
        break
