import torch.nn.functional as F
import torch.nn.init as init
import torch.nn as nn
import torch
import math


class MLP(nn.Module):
    def __init__(self, mlp=[1024, 512, 512], activate='leaky', normal='ln', output='none', bias=True):
        super(MLP, self).__init__()
        layers = []
        for i in range(len(mlp)-1):
            layers.append(nn.Linear(mlp[i], mlp[i+1], bias=bias))
            if i != len(mlp) - 2:
                # normal
                if normal == 'bn':
                    layers.append(nn.BatchNorm1d(mlp[i+1]))
                elif normal == 'ln':
                    layers.append(nn.LayerNorm(mlp[i+1]))
                elif normal == 'none':
                    pass
                # activation
                if activate == 'leaky':
                    layers.append(nn.LeakyReLU())
                elif activate == 'relu':
                    layers.append(nn.ReLU())
                elif activate == 'tanh':
                    layers.append(nn.Tanh())
            if i == len(mlp)-2:
                if output == 'log_sm':
                    layers.append(nn.LogSoftmax(dim=-1))
                elif output == 'sm':
                    layers.append(nn.Softmax(dim=-1))
                elif output == 'sigmoid':
                    layers.append(nn.Sigmoid())
                elif output == 'none':
                    pass
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        y = self.layers(x)
        return y


class SemanticGraph():
    def __init__(self, x, edge_index):
        self.x = x
        self.h = torch.clone(x)
        self.h_outputs = []
        self.edge_index = edge_index

    def reset_graph(self):
        self.h_outputs = []


class GraphAttnEncoder(nn.Module):
    def __init__(self, ndim=512, nlayer=3):
        super(GraphAttnEncoder, self).__init__()
        self.nlayer = nlayer
        self.ndim = ndim
        self.sgconv = nn.ModuleList([AttnConv(ndim=self.ndim) for i in range(self.nlayer)])
        self.LN = nn.LayerNorm(self.ndim)

    def forward(self, G: SemanticGraph):
        s_idx, o_idx = G.edge_index[0, :].contiguous(), G.edge_index[1, :].contiguous()
        for i in range(self.nlayer):
            G = self.sgconv[i](G, s_idx, o_idx)
        G.h_outputs = torch.cat(G.h_outputs, dim=0)
        G.h = self.LN(G.h_outputs.sum(dim=0))
        return G.h


class AttnConv(nn.Module):
    def __init__(self, ndim=512):
        super(AttnConv, self).__init__()
        self.ndim = ndim
        self.edge_encoder = MLP(mlp=[self.ndim, self.ndim, self.ndim])
        self.edge_msg_encoder = MLP(mlp=[self.ndim, self.ndim, self.ndim])
        self.attn_encoder = MLP(mlp=[self.ndim, self.ndim, 1])
        self.sigmoid = nn.Sigmoid()
        self.node_GRU = nn.GRUCell(self.ndim, self.ndim)

    def forward(self, G: SemanticGraph, s_idx, o_idx):
        n_ins = G.x.shape[0]
        Hs, Ho = G.h[s_idx], G.h[o_idx]
        msg_node = self.message(Hs, Ho, o_idx, n_ins)
        G.h = self.node_GRU(msg_node, G.h)
        G.h_outputs.append(G.h.view(1, -1, self.ndim))
        return G

    def message(self, Hs, Ho, o_idx, n_ins):
        ''' calculate message from subject to object
        '''
        msg_p = self.message_node(Hs, Ho)
        attn_w = self.attn_weight(Hs, Ho)
        msg_p = attn_w * msg_p
        msg_node = self.attn_pooling(msg_p, o_idx, n_ins)
        return msg_node

    def message_node(self, Hs, Ho):
        msg_node = F.leaky_relu(Hs - Ho)
        return msg_node

    def attn_weight(self, Hs, Ho):
        attn_w = self.attn_encoder(Hs - Ho)
        attn_w = self.sigmoid(attn_w)
        return attn_w

    def attn_pooling(self, msg, o_idx, n_ins):
        msg_pooled = torch.zeros(n_ins, self.ndim).cuda()
        msg_pooled = msg_pooled.scatter_add(0, o_idx.view(-1, 1).expand_as(msg), msg)
        return msg_pooled


class FCNN(nn.Module):
    def __init__(self, in_dim, out_dim, hidden_dim):
        super(FCNN, self).__init__()
        self.network = nn.Sequential(
            nn.Linear(in_dim, hidden_dim),
            nn.Tanh(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.Tanh(),
            nn.Linear(hidden_dim, out_dim),
        )
        # self.network = MLP(mlp=[in_dim, hidden_dim, out_dim], activate='leaky', normal='ln')

    def forward(self, x):
        return self.network(x)


class MAF(nn.Module):
    def __init__(self, dim, hidden_dim=16, base_network=FCNN):
        super(MAF, self).__init__()
        self.dim = dim
        self.layers = nn.ModuleList()
        self.initial_param = nn.Parameter(torch.Tensor(2))
        for i in range(1, dim):
            self.layers += [base_network(i, 2, hidden_dim)]
        self.reset_parameters()

    def reset_parameters(self):
        init.uniform_(self.initial_param, -math.sqrt(0.5), math.sqrt(0.5))

    def forward(self, x):
        z = torch.zeros_like(x)
        log_det = torch.zeros(z.shape[0]).cuda()
        for i in range(self.dim):
            if i == 0:
                mu, alpha = self.initial_param[0], self.initial_param[1]
            else:
                out = self.layers[i - 1](x[:, :i])
                mu, alpha = out[:, 0], out[:, 1]
            z[:, i] = (x[:, i] - mu) / torch.exp(alpha)
            log_det -= alpha
        return z.flip(dims=(1,)), log_det

    def inverse(self, z):
        x = torch.zeros_like(z)
        log_det = torch.zeros(z.shape[0]).cuda()
        z = z.flip(dims=(1,))
        for i in range(self.dim):
            if i == 0:
                mu, alpha = self.initial_param[0], self.initial_param[1]
            else:
                out = self.layers[i - 1](x[:, :i])
                mu, alpha = out[:, 0], out[:, 1]
            x[:, i] = mu + torch.exp(alpha) * z[:, i]
            log_det += alpha
        return x, log_det


class ConditionalMAF(nn.Module):
    def __init__(self, dim, condition_dim, hidden_dim=16, base_network=FCNN):
        super(ConditionalMAF, self).__init__()
        self.dim = dim
        self.condition_dim = condition_dim
        self.layers = nn.ModuleList()
        self.initial_param = nn.Parameter(torch.Tensor(2))
        for i in range(1, dim):
            self.layers += [base_network(i + self.condition_dim, 2, hidden_dim)]
        self.reset_parameters()

    def reset_parameters(self):
        init.uniform_(self.initial_param, -math.sqrt(0.5), math.sqrt(0.5))

    def forward(self, x, condition):
        z = torch.zeros_like(x)
        log_det = torch.zeros(z.shape[0]).cuda()
        for i in range(self.dim):
            if i == 0:
                mu, alpha = self.initial_param[0], self.initial_param[1]
            else:
                conditional_x = torch.cat([x[:, :i], condition], dim=1)
                out = self.layers[i - 1](conditional_x)
                mu, alpha = out[:, 0], out[:, 1]
            z[:, i] = (x[:, i] - mu) / torch.exp(alpha)
            log_det -= alpha
        return z.flip(dims=(1,)), log_det

    def inverse(self, z, condition):
        x = torch.zeros_like(z)
        log_det = torch.zeros(z.shape[0]).cuda()
        z = z.flip(dims=(1,))
        for i in range(self.dim):
            if i == 0:
                mu, alpha = self.initial_param[0], self.initial_param[1]
            else:
                conditional_x = torch.cat([x[:, :i], condition], dim=1)
                out = self.layers[i - 1](conditional_x)
                mu, alpha = out[:, 0], out[:, 1]
            x[:, i] = mu + torch.exp(alpha) * z[:, i]
            log_det += alpha
        return x, log_det


class NormalizingFlowModel(nn.Module):
    def __init__(self, prior, flows):
        super().__init__()
        self.prior = prior
        self.flows = nn.ModuleList(flows)

    def forward(self, x):
        m, _ = x.shape
        log_det = torch.zeros(m).cuda()
        for flow in self.flows:
            x, ld = flow.forward(x)
            log_det += ld
        z, prior_logprob = x, self.prior.log_prob(x)
        return z, prior_logprob, log_det

    def inverse(self, z):
        m, _ = z.shape
        log_det = torch.zeros(m).cuda()
        for flow in self.flows[::-1]:
            z, ld = flow.inverse(z)
            log_det += ld
        x = z
        return x, log_det

    def sample(self, n_samples):
        z = self.prior.sample((n_samples,))
        x, _ = self.inverse(z)
        return x


class ConditionalNormalizingFlowModel(nn.Module):
    def __init__(self, prior, flows):
        super(ConditionalNormalizingFlowModel, self).__init__()
        self.prior = prior
        self.flows = nn.ModuleList(flows)

    def forward(self, x, condition):
        m, _ = x.shape
        log_det = torch.zeros(m).cuda()
        for flow in self.flows:
            x, ld = flow.forward(x, condition)
            log_det += ld
        z, prior_logprob = x, self.prior.log_prob(x)
        return z, prior_logprob, log_det

    def inverse(self, z, condition):
        m, _ = z.shape
        log_det = torch.zeros(m).cuda()
        for flow in self.flows[::-1]:
            z, ld = flow.inverse(z, condition)
            log_det += ld
        x = z
        return x, log_det

    def sample(self, n_samples, condition):
        z = self.prior.sample((n_samples,))
        x, _ = self.inverse(z, condition)
        return x


if __name__ == '__main__':
    print(MLP(mlp=[128, 128, 256], activate='leaky', normal='none', output='none').cuda())
