# import pathmagic  # noqa
from data.data_loader import ThreedFrontDataset, ThreedFrontGroupDataset
from scipy.spatial.transform import Rotation
from scrtips_utils import CheckPointUtils, OptimizationModel
from scrtips_utils import VisualizerUtils
import matplotlib.pyplot as plt  # noqa
from sklearn.manifold import TSNE  # noqa
from tqdm import tqdm  # noqa
import numpy as np  # noqa
import importlib
import torch
import vedo  # noqa
import time  # noqa
import math
import json  # noqa
import sys
import os

# import json
# import data.visualizer as visualizer


PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))
EPS = 1e-5


def build_optimizer(network, lr=0.0001, m=0.9):
    optimizer = torch.optim.SGD(network.parameters(), lr=lr, momentum=m)
    return optimizer


class UserInterface():
    def __init__(self):
        self.group_dict = {
            "Sleeping":         {"existence": 1, "complexity": 5},  # 0 [1, 2, 3, 4, 5, 6]
            "Relaxing":         {"existence": 0, "complexity": 0},  # 1 [1, 2, 3, 4]
            "BedWatchingTV":    {"existence": 0, "complexity": 0},  # 2 [1, 2, 3, 5]
            "Lighting":         {"existence": 0, "complexity": 0},  # 3 [1, 2, 3, 4]
            "Dressing":         {"existence": 1, "complexity": 2},  # 4 [1, 2, 3]
            "ClothStorage":     {"existence": 2, "complexity": 1},  # 5 [1, 2, 3, 4]
            "BookStorage":      {"existence": 0, "complexity": 0},  # 6 [1, 2, 3, 4, 5, 6]
            "ShelfStorage":     {"existence": 0, "complexity": 0},  # 7 [1, 2, 3]
            "KitchenStorage":   {"existence": 0, "complexity": 0},  # 8 [1, 2, 3, 4]
            "WatchingTV":       {"existence": 0, "complexity": 0},  # 9 [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
            "Working":          {"existence": 0, "complexity": 0},  # 10 [1, 2, 3, 4, 5, 6, 8]
            "Eating":           {"existence": 0, "complexity": 0}   # 11 [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        }

    def process_complexity_map(self):
        # load complexity map
        with open('./data/complexity_map.json', 'r') as f:
            complexity_map = json.load(f)[self.room_type]
        f.close()
        print(complexity_map)

    def parse_user_intent(self):
        key_list = list(self.group_dict.keys())
        group_lab = []
        group_complexity = []
        for i in range(len(key_list)):
            for _ in range(self.group_dict[key_list[i]]["existence"]):
                group_lab.append(i)
                group_complexity.append(self.group_dict[key_list[i]]["complexity"])
        group_lab = torch.Tensor(group_lab)
        group_complexity = torch.Tensor(group_complexity)
        print(group_lab, group_complexity)
        # merge 0, 2
        idx_0, idx_2 = group_lab == 0, group_lab == 2
        group_lab_list, group_complexity_list = [], []
        if idx_0.sum() > 0 and idx_2.sum() > 0:
            for i in range(group_lab.shape[0]):
                if group_lab[i] == 0:
                    continue
                else:
                    group_lab_list.append(group_lab[i])
                    group_complexity_list.append(group_complexity[i])
                # if group_lab[i] == 0:
                #     group_lab_list.append(group_lab[i])
                #     group_complexity_list.append(group_complexity[i])
                # else:
                #     continue
        group_lab = torch.Tensor(group_lab).long().cuda()
        group_complexity = torch.Tensor(group_complexity).long().cuda()
        return group_lab, group_complexity


class Propose_Complete_Pipeline():
    def __init__(self, room_type):
        # dataset
        self.room_type = room_type
        # self.SCENE_SET = ThreedFrontDataset(room_type=self.room_type, split='test', as_user_intent=True)
        self.SCENE_SET = ThreedFrontDataset(room_type='bedroom', split='train', as_user_intent=True)
        self.GROUP_SET = ThreedFrontGroupDataset(room_type=self.room_type)
        # parameters
        # self.sample_idx = 328  # 6, 87, 116, 133, 328, 638, 663, 1257 (big rect)
        self.n_proposal = 50
        self.th_delta = 0.125
        # trained models
        self.complete_model = CompleteGroup(self.GROUP_SET, delta=self.th_delta, mode='complete', room_type=self.room_type)
        if self.room_type == 'bedroom':
            self.proposal_model = ProposeNetModel('model_proposenet_bedroom', 'model_proposenet', 'model_timestamp')
            self.complete_model.load_network('model_completenet_vae_bedroom', 'model_completenet_vae', 'model_timestamp')
        elif self.room_type == 'diningroom':
            self.proposal_model = ProposeNetModel('model_proposenet_diningroom', 'model_proposenet', 'model_timestamp')
            self.complete_model.load_network('model_completenet_vae_diningroom', 'model_completenet_vae', 'model_timestamp')
        elif self.room_type == 'library':
            self.proposal_model = ProposeNetModel('model_proposenet_library', 'model_proposenet', 'model_timestamp')
            self.complete_model.load_network('model_completenet_vae_library', 'model_completenet_vae', 'model_timestamp')
        elif self.room_type == 'livingroom':
            self.proposal_model = ProposeNetModel('model_proposenet_livingroom', 'model_proposenet', 'model_timestamp')
            self.complete_model.load_network('model_completenet_vae_livingroom', 'model_completenet_vae', 'model_timestamp')
        elif self.room_type == 'all':
            self.proposal_model = ProposeNetModel('model_proposenet_all', 'model_proposenet', 'model_timestamp')
            self.complete_model.load_network('model_completenet_vae_all', 'model_completenet_vae', 'model_timestamp')
        # util models
        self.user_interface = UserInterface()
        self.optimizer = OptimizationModel()
        self.cpts_utils = CheckPointUtils(n_proposal=self.n_proposal)
        self.visualizer_utils = VisualizerUtils(room_type=self.room_type)
        # load complexity map
        with open('./data/complexity_map.json', 'r') as f:
            self.complexity_map = json.load(f)[self.room_type]
        f.close()

    def parse_data_as_user_intent(self, sample_idx):
        _, _, group_lab, _, N, _, group_complexity = self.SCENE_SET.__getitem__(sample_idx)
        group_lab = group_lab[:N]
        # merge 0, 2
        idx_0, idx_2 = group_lab == 0, group_lab == 2
        if idx_0.sum() > 0 and idx_2.sum() > 0:
            group_lab_list, group_complexity_list = [], []
            for i in range(group_lab.shape[0]):
                if group_lab[i] == 0:
                    continue
                else:
                    group_lab_list.append(group_lab[i])
                    group_complexity_list.append(group_complexity[i])
            group_lab = torch.Tensor(group_lab_list).long().cuda()
            group_complexity = torch.Tensor(group_complexity_list).long().cuda()
            return group_lab, group_complexity
        return torch.Tensor(group_lab).long().cuda(), torch.Tensor(group_complexity).long().cuda()

    def process_user_intent_synthetic_scene_opt(self, sample_idx, img_idx=0, mode='user_intent', json_output=True):  # mode = dataset / user_intent
        _, _, _, floor_mask, _, floor_scales, _ = self.SCENE_SET.__getitem__(sample_idx)
        _, floor_mesh, _ = self.SCENE_SET.visualize_scene(sample_idx, return_vis=True)  # vis_data
        floor_cpts = floor_check_points(floor_mask)
        if mode == 'dataset_intent':
            group_lab, group_complexity = self.parse_data_as_user_intent(sample_idx)
        elif mode == 'user_intent':
            group_lab, group_complexity = self.user_interface.parse_user_intent()
        self.visualizer_utils.add_floor(floor_mesh, floor_cpts)
        with torch.no_grad():
            pred_posrot, pred_label = self.proposal_model.propose(group_lab, self.n_proposal, floor_scales)
            bbox_ori_list, cpts_ori_list, group_lab, group_complexity = self.complete_model.complete(group_lab, group_complexity)
        bbox_selected_list, cpts_selected_list, posrot_selected = [], [], []
        for i in range(group_lab.shape[0]):
            posrot = pred_posrot[i*self.n_proposal: (i+1)*self.n_proposal]
            bbox_trans = self.cpts_utils.transform_batch_group(posrot[:, 0], posrot[:, 2], posrot[:, 3], bbox_ori_list[i], posrot.shape[0])
            cpts_trans = self.cpts_utils.transform_batch_cpts(posrot[:, 0], posrot[:, 2], posrot[:, 3], cpts_ori_list[i], posrot.shape[0])
            best_idx = self.cpts_utils.get_best_idx(bbox_trans, cpts_trans, posrot[:, 3], floor_cpts, self.th_delta, self.n_proposal, group_lab[i])
            bbox_selected_list.append(bbox_trans[best_idx].squeeze(0))
            cpts_selected_list.append(cpts_trans[best_idx].squeeze(0))
            posrot_selected.append(posrot[best_idx])
            if not json_output:
                self.visualizer_utils.add_best_group(bbox_trans[best_idx], cpts_trans[best_idx], group_lab[i])
        self.cpts_utils.reset()
        posrot_selected = torch.cat(posrot_selected, dim=0)
        # optimization
        bbox_opt_list = self.optimizer.run(posrot_selected, cpts_ori_list, bbox_ori_list, floor_cpts, self.th_delta, group_lab, max_iter=500)
        if not json_output:
            for i in range(len(bbox_opt_list)):
                self.visualizer_utils.add_opt_group(bbox_opt_list[i])
            self.visualizer_utils.show_opt()
        else:
            root = 'save_dir'
            with open(os.path.join(root, f'{sample_idx}_{img_idx}.json'), 'w') as f:
                json_dict = {}
                json_dict['data'] = [bbox.cpu().numpy().tolist() for bbox in bbox_opt_list]
                f.write(json.dumps(json_dict))
            f.close()

    def process_testset_synthetic_scene_opt(self, sample_idx, img_idx=None, mode='dataset_intent', json_output=True):  # mode = dataset / user_intent
        _, _, _, floor_mask, _, floor_scales, _ = self.SCENE_SET.__getitem__(sample_idx)
        _, floor_mesh, _ = self.SCENE_SET.visualize_scene(sample_idx, return_vis=True)  # vis_data
        floor_cpts = floor_check_points(floor_mask)
        if mode == 'dataset_intent':
            group_lab, group_complexity = self.parse_data_as_user_intent(sample_idx)
        elif mode == 'user_intent':
            group_lab, group_complexity = self.user_interface.parse_user_intent()
        self.visualizer_utils.add_floor(floor_mesh, floor_cpts)
        with torch.no_grad():
            pred_posrot, pred_label = self.proposal_model.propose(group_lab, self.n_proposal, floor_scales)
            bbox_ori_list, cpts_ori_list, group_lab, group_complexity = self.complete_model.complete(group_lab, group_complexity)
        # select the best proposal
        bbox_selected_list, cpts_selected_list, posrot_selected = [], [], []
        for i in range(group_lab.shape[0]):
            posrot = pred_posrot[i*self.n_proposal: (i+1)*self.n_proposal]
            bbox_trans = self.cpts_utils.transform_batch_group(posrot[:, 0], posrot[:, 2], posrot[:, 3], bbox_ori_list[i], posrot.shape[0])
            cpts_trans = self.cpts_utils.transform_batch_cpts(posrot[:, 0], posrot[:, 2], posrot[:, 3], cpts_ori_list[i], posrot.shape[0])
            best_idx = self.cpts_utils.get_best_idx(bbox_trans, cpts_trans, posrot[:, 3], floor_cpts, self.th_delta, self.n_proposal, group_lab[i])
            bbox_selected_list.append(bbox_trans[best_idx].squeeze(0))
            cpts_selected_list.append(cpts_trans[best_idx].squeeze(0))
            posrot_selected.append(posrot[best_idx])
            if not json_output:
                self.visualizer_utils.add_best_group(bbox_trans[best_idx], cpts_trans[best_idx], group_lab[i])
        self.cpts_utils.reset()
        posrot_selected = torch.cat(posrot_selected, dim=0)
        # optimization
        bbox_opt_list = self.optimizer.run(posrot_selected, cpts_ori_list, bbox_ori_list, floor_cpts, self.th_delta, group_lab, floor_mesh, max_iter=500)
        if not json_output:
            for i in range(len(bbox_opt_list)):
                self.visualizer_utils.add_opt_group(bbox_opt_list[i])
            self.visualizer_utils.show_opt()
        else:
            root = 'save_dir'
            with open(os.path.join(root, f'{sample_idx}_{img_idx}.json'), 'w') as f:
                json_dict = {}
                json_dict['data'] = [bbox.cpu().numpy().tolist() for bbox in bbox_opt_list]
                f.write(json.dumps(json_dict))
            f.close()
            root = 'save_dir'
            with open(os.path.join(root, f'{sample_idx}_{img_idx}.json'), 'w') as f:
                json_dict = {}
                json_dict['data'] = [bbox.cpu().numpy().tolist() for bbox in bbox_selected_list]
                f.write(json.dumps(json_dict))
            f.close()

    def time_statistic(self, sample_idx, mode='dataset_intent'):  # mode = dataset / user_intent
        _, _, _, floor_mask, _, floor_scales, _ = self.SCENE_SET.__getitem__(sample_idx)
        _, floor_mesh, _ = self.SCENE_SET.visualize_scene(sample_idx, return_vis=True)  # vis_data
        time_start = time.clock()
        floor_cpts = floor_check_points(floor_mask)
        if mode == 'dataset_intent':
            group_lab, group_complexity = self.parse_data_as_user_intent(sample_idx)
        elif mode == 'user_intent':
            group_lab, group_complexity = self.user_interface.parse_user_intent()
        with torch.no_grad():
            pred_posrot, pred_label = self.proposal_model.propose(group_lab, self.n_proposal, floor_scales)
            bbox_ori_list, cpts_ori_list, group_lab, group_complexity = self.complete_model.complete(group_lab, group_complexity)
        # select the best proposal
        bbox_selected_list, cpts_selected_list, posrot_selected = [], [], []
        for i in range(group_lab.shape[0]):
            posrot = pred_posrot[i*self.n_proposal: (i+1)*self.n_proposal]
            bbox_trans = self.cpts_utils.transform_batch_group(posrot[:, 0], posrot[:, 2], posrot[:, 3], bbox_ori_list[i], posrot.shape[0])
            cpts_trans = self.cpts_utils.transform_batch_cpts(posrot[:, 0], posrot[:, 2], posrot[:, 3], cpts_ori_list[i], posrot.shape[0])
            best_idx = self.cpts_utils.get_best_idx(bbox_trans, cpts_trans, posrot[:, 3], floor_cpts, self.th_delta, self.n_proposal, group_lab[i])
            bbox_selected_list.append(bbox_trans[best_idx].squeeze(0))
            cpts_selected_list.append(cpts_trans[best_idx].squeeze(0))
            posrot_selected.append(posrot[best_idx])
        posrot_selected = torch.cat(posrot_selected, dim=0)
        # optimization
        self.optimizer.run_time_statitics(posrot_selected, cpts_ori_list, bbox_ori_list, floor_cpts, self.th_delta, group_lab, floor_mesh, max_iter=500)
        time_end = time.clock()
        print(f'time: {time_end - time_start}')


class ProposeNetModel():
    def __init__(self, dir_name, model_name, time_stamp):
        self.network = self.load_network(dir_name, model_name, time_stamp)

    def load_network(self, dir_name, model_name, time_stamp):
        address = os.path.join(PROJECT_DIR, 'log', dir_name, time_stamp)
        dict_address = os.path.join(address, 'checkpoints/best_model.pth')
        sys.path.append(address)
        MODEL = importlib.import_module(model_name)
        network = MODEL.get_model()
        checkpoint = torch.load(dict_address)
        network.load_state_dict(checkpoint['model_state_dict'])
        network = network.cuda().eval()
        return network

    def propose(self, group_lab, n_proposal, floor_scales):
        pred_posrot, pred_label = self.network.sample_single_group(group_lab, n_proposal)
        pred_posrot = torch.clamp(pred_posrot, min=-1, max=1)
        pred_posrot[:, 0] = pred_posrot[:, 0] * floor_scales[1]
        pred_posrot[:, 2] = pred_posrot[:, 2] * floor_scales[0]
        return pred_posrot, pred_label


class CompleteGroup():
    def __init__(self, dataset: ThreedFrontGroupDataset, delta=0.1, mode='dataset', room_type='bedroom'):
        self.dataset = dataset
        self.delta = delta
        self.mode = mode
        self.network = None
        with open('./data/complexity_map.json', 'r') as f:
            self.complexity_map = json.load(f)[room_type]
        f.close()

    def load_network(self, dir_name, model_name, time_stamp):
        address = os.path.join(PROJECT_DIR, 'log', dir_name, time_stamp)
        dict_address = os.path.join(address, 'checkpoints/best_model.pth')
        # dict_address = os.path.join(address, 'checkpoints/best_model_950.pth')
        sys.path.append(address)
        MODEL = importlib.import_module(model_name)
        network = MODEL.get_model()
        checkpoint = torch.load(dict_address)
        network.load_state_dict(checkpoint['model_state_dict'])
        network = network.cuda().eval()
        self.network = network

    def complete(self, group_lab, group_complexity):
        bbox_list = []
        cpts_list = []
        # get_group_bbox
        for i in range(group_lab.shape[0]):
            bbox_list.append(self.get_group_bbox(group_lab[i], group_complexity[i]))
        # force on the floor
        whitelist = [4, 10, 13, 21]
        for i in range(len(bbox_list)):
            bbox = bbox_list[i]
            for j in range(bbox.shape[0]):
                if bbox[j, 7] not in whitelist:
                    bbox[j, 1] = bbox[j, 5] * 0.5 + 0.5
            bbox_list[i] = bbox
        # generate cpts
        for i in range(len(bbox_list)):
            cpts_in = self.group_check_points_with_functional_area(bbox_list[i], group_lab[i], self.delta)
            cpts_list.append(cpts_in)
        # order by size
        sizes, bbox_order_list, cpts_order_list = [], [], []
        for i in range(len(bbox_list)):
            sizes.append(cpts_list[i].shape[0])
        sizes = torch.Tensor(sizes).long().cuda()
        idx_sort = torch.argsort(sizes, descending=True)
        for i in range(len(bbox_list)):
            bbox_order_list.append(bbox_list[idx_sort[i]])
            cpts_order_list.append(cpts_list[idx_sort[i]])
        group_lab = group_lab[idx_sort]
        group_complexity = group_complexity[idx_sort]
        return bbox_order_list, cpts_order_list, group_lab, group_complexity

    def get_group_bbox(self, lab, nins):
        if self.mode == 'dataset':
            group_bbox = torch.Tensor(self.dataset.get_group(lab)).cuda()
            return group_bbox
        elif self.mode == 'complete':
            group_idx = torch.Tensor([lab]).long().cuda()
            nins = torch.Tensor([nins]).long().cuda()
            available_complexity = self.complexity_map[group_idx.item()]
            if nins.item() not in available_complexity:
                nins = torch.Tensor([available_complexity[-1]]).long().cuda()
            pred_output = self.network.sample(group_idx[0], nins[0])
            pred_group_lab = group_idx.repeat(nins).view(-1, 1).contiguous()
            pred_bbox = self.network.continuous_output(pred_output, type='pred')
            pred_bbox = torch.cat([pred_bbox, pred_output['class'].topk(k=1).indices.view(-1, 1), pred_group_lab], dim=-1)
            return pred_bbox

    def group_check_points(self, group_bbox, delta):
        bbox = self.corner_points(group_bbox)
        maxs = bbox.max(dim=0).values
        mins = bbox.min(dim=0).values
        offset = (maxs + mins) * 0.5
        offset[1] = 0
        scale = maxs - mins
        N_x = torch.round(scale[0] / delta)
        N_y = torch.Tensor([0])
        N_z = torch.round(scale[2] / delta)
        x, y, z = torch.meshgrid(torch.arange(N_x), N_y, torch.arange(N_z))
        cpts_in = (torch.stack([x, y, z], 3).cuda() * delta).view(-1, 3).contiguous()
        cpts_in = cpts_in - torch.mean(cpts_in, dim=0) + offset
        return cpts_in

    def group_check_points_with_functional_area(self, group_bbox, group_label, delta):
        func_area_list = [5, 6, 7, 8]
        group_label = group_label.item()
        bbox = self.corner_points(group_bbox)
        maxs = bbox.max(dim=0).values
        mins = bbox.min(dim=0).values
        offset = (maxs + mins) * 0.5
        offset[1] = 0
        scale = maxs - mins
        N_x = torch.round(scale[0] / delta)
        N_y = torch.Tensor([0])
        N_z = torch.round(scale[2] / delta)
        # inner check points
        x, y, z = torch.meshgrid(torch.arange(N_x), N_y, torch.arange(N_z))
        cpts_in = (torch.stack([x, y, z], 3).cuda() * delta).view(-1, 3).contiguous()
        cpts_in = cpts_in - torch.mean(cpts_in, dim=0) + offset
        if group_label in func_area_list:
            # functional area check points
            N_zo = torch.Tensor([0, 1, 2])
            xo, yo, zo = torch.meshgrid(torch.arange(N_x), N_y, N_zo)
            cpts_func = (torch.stack([xo, yo, zo], 3).cuda() * delta).view(-1, 3).contiguous()
            cpts_func = cpts_func - torch.mean(cpts_func, dim=0) + offset + torch.Tensor([0, 0, (scale[2] + N_zo.shape[0]*delta) * 0.5]).cuda()
            cpts = torch.cat([cpts_in, cpts_func], dim=0)
            return cpts
        else:
            return cpts_in

    def corner_points(self, group_bbox):
        pos = group_bbox[:, 0:3] * 4
        rot = group_bbox[:, 3] * 180 + 180
        scale = group_bbox[:, 4:7] * 4 + 4
        bbox = []
        for i in range(group_bbox.shape[0]):
            bbox.append(self.transform_bbox(self.unit_bbox(), pos[i], rot[i], scale[i]))
        bbox = torch.cat(bbox, dim=0)
        return bbox

    def transform_bbox(self, unit_bbox, pos, rot, scale):
        pts = unit_bbox * scale
        rot = rot.cpu().numpy()
        rotMatrix = Rotation.from_rotvec((rot / 360.) * 2 * math.pi * np.array([0, 1, 0])).as_matrix()
        R = torch.Tensor(np.array(rotMatrix)).cuda()
        pts = torch.matmul(R, pts.transpose(0, 1)).transpose(0, 1)
        pts = pts + pos
        return pts

    def unit_bbox(self):
        return torch.Tensor([
                [0.5, -0.5, 0.5],
                [0.5, -0.5, -0.5],
                [-0.5, -0.5, -0.5],
                [-0.5, -0.5, 0.5],
                [0.5, 0.5, 0.5],
                [0.5, 0.5, -0.5],
                [-0.5, 0.5, -0.5],
                [-0.5, 0.5, 0.5]
            ]).cuda()


def floor_check_points(floor_mask):
    pts = []
    for i in range(floor_mask.shape[0]):
        for j in range(floor_mask.shape[1]):
            if i > 0 and i < floor_mask.shape[0] - 1 and j > 0 and j < floor_mask.shape[1] - 1:
                if floor_mask[i, j] != 0 and check_around(floor_mask, i, j):
                    pts.append([j, 0, -i])
    pts = torch.Tensor(pts).cuda()
    pts = (pts - torch.Tensor([31.5, 0, -31.5]).cuda()) / 8
    return pts


def check_around(floor_mask, i, j):
    if floor_mask[i+1, j] == 0 or floor_mask[i-1, j] == 0 or floor_mask[i, j-1] == 0 or floor_mask[i, j+1] == 0 or \
       floor_mask[i+1, j+1] == 0 or floor_mask[i-1, j+1] == 0 or floor_mask[i+1, j-1] == 0 or floor_mask[i-1, j-1] == 0:
        return False
    else:
        return True


def generate_top_down_json(room_type='bedroom'):
    start = 75
    propose_and_complete = Propose_Complete_Pipeline(room_type)
    print(len(propose_and_complete.SCENE_SET))
    for i in tqdm(range(start, len(propose_and_complete.SCENE_SET))):
        for j in range(6):
            propose_and_complete.process_testset_synthetic_scene_json(i, j, mode='dataset_intent', json_output=True)


def generate_json_for_KL(room_type='bedroom'):
    start = 0
    propose_and_complete = Propose_Complete_Pipeline(room_type)
    print(len(propose_and_complete.SCENE_SET))
    for i in tqdm(range(start, len(propose_and_complete.SCENE_SET))):
        for j in range(6):
            propose_and_complete.process_testset_synthetic_scene_kl(i, j, mode='dataset_intent', json_output=True)

