import pathmagic  # noqa
from data.data_loader import ThreedFrontDataset
import os
import sys
import argparse
import logging
import time
import datetime
import importlib
import shutil
from tqdm import tqdm
from pathlib import Path
import numpy as np
import torch


# Project directory
PROJECT_DIR = os.path.dirname(__file__)
sys.path.append(os.path.join(PROJECT_DIR, 'models'))
FILE = __file__
os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'


# Arguments declearation
def parse_args():
    parser = argparse.ArgumentParser('Model')
    parser.add_argument('--model', type=str, default='model_proposenet', help='model name [default: proposenet]')
    parser.add_argument('--room_type', type=str, default='bedroom', help='all, bedroom, diningroom, library, livingroom')
    parser.add_argument('--epoch',  default=1000, type=int, help='Epoch to run [default: 100]')
    parser.add_argument('--batch_size',  default=16, type=int, help='Batch size to run')
    parser.add_argument('--gpu', type=str, default='0', help='GPU to use [default: GPU 0]')
    parser.add_argument('--log_dir', type=str, default=None, help='Log path [default: None]')

    parser.add_argument('--optimizer', type=str, default='Adam', help='Adam or SGD [default: Adam]')
    parser.add_argument('--learning_rate', default=1e-3, type=float, help='Initial learning rate [default: 0.001]')
    parser.add_argument('--decay_rate', type=float, default=1e-4, help='weight decay [default: 1e-4]')
    parser.add_argument('--step_size', type=int,  default=50, help='Decay step for lr decay [default: every 10 epochs]')
    parser.add_argument('--lr_decay', type=float,  default=0.90, help='Decay rate for lr decay [default: 0.7]')

    return parser.parse_args()


def main(args):
    def log_string(str):
        logger.info(str)
        print(str)

    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu  # cuda
    # torch.backends.cudnn.enabled = False  # disable cudnn for CUDNN_NOT_SUPPORT, may not contiguous
    # ---------------- Create log dir -------------------
    timestr = str(datetime.datetime.now().strftime('%Y-%m-%d_%H-%M'))
    experiment_dir = Path('./log/')
    experiment_dir.mkdir(exist_ok=True)
    experiment_dir = experiment_dir.joinpath(f'{args.model}_{args.room_type}')
    experiment_dir.mkdir(exist_ok=True)
    if args.log_dir is None:
        experiment_dir = experiment_dir.joinpath(timestr)
    else:
        experiment_dir = experiment_dir.joinpath(args.log_dir)
    experiment_dir.mkdir(exist_ok=True)
    checkpoints_dir = experiment_dir.joinpath('checkpoints/')
    checkpoints_dir.mkdir(exist_ok=True)
    log_dir = experiment_dir.joinpath('logs/')
    log_dir.mkdir(exist_ok=True)

    args = parse_args()
    logger = logging.getLogger("Model")
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler('%s/%s.txt' % (log_dir, args.model))
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    log_string('PARAMETER ...')
    log_string(args)

    # ---------------- Training set --------------------
    TRAINING_SET = ThreedFrontDataset(room_type=args.room_type, split='train', rot_aug=True)
    trainDataLoader = torch.utils.data.DataLoader(TRAINING_SET, batch_size=args.batch_size, shuffle=True, num_workers=0, pin_memory=True,
                                                  drop_last=True, worker_init_fn=lambda x: np.random.seed(x+int(time.time())))
    # ---------------- Log dataset info --------------------
    log_string("The number of training data is: %d" % len(TRAINING_SET))

    # ---------------- Config network model --------------------
    MODEL = importlib.import_module(args.model)
    shutil.copy(os.path.join(PROJECT_DIR, 'models', f'{args.model}.py'), str(experiment_dir))
    shutil.copy(os.path.join(PROJECT_DIR, 'models', 'model.py'), str(experiment_dir))
    shutil.copy(os.path.join(PROJECT_DIR, 'models', 'utils.py'), str(experiment_dir))
    shutil.copy(os.path.join(FILE), str(experiment_dir))

    network = MODEL.get_model().cuda()    # cuda
    criterion = MODEL.get_loss().cuda()   # cuda

    def weights_init(m):
        classname = m.__class__.__name__
        if classname.find('Conv2d') != -1:
            torch.nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                torch.nn.init.constant_(m.bias.data, 0.0)
        elif classname.find('Linear') != -1:
            torch.nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                torch.nn.init.constant_(m.bias.data, 0.0)
        elif classname.find('Conv1d') != -1:
            torch.nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                torch.nn.init.constant_(m.bias.data, 0.0)

    network = network.apply(weights_init)

    # training based on old model
    # address = address = os.path.join(PROJECT_DIR, 'log/model_propose_autoregressive/2022-10-13_22-49')
    # dict_address = os.path.join(address, 'checkpoints/best_model.pth')
    # sys.path.append(address)
    # checkpoint = torch.load(dict_address)
    # network.load_state_dict(checkpoint['model_state_dict'])
    # network = network.cuda()
    start_epoch = 0

    def build_optimizer(network, args_opt='Adam', args_lr=0.001, args_weight_decay=0.0001):
        if args_opt == 'Adam':
            optimizer = torch.optim.Adam(
                network.parameters(),
                lr=args_lr,
                betas=(0.9, 0.999),
                eps=1e-08,
                weight_decay=args_weight_decay
                )
        else:
            optimizer = torch.optim.SGD(network.parameters(), lr=args_lr, momentum=0.9)
        return optimizer

    optimizer = build_optimizer(network,
                                args.optimizer,
                                args.learning_rate,
                                args.decay_rate)

    def bn_momentum_adjust(m, momentum):
        if isinstance(m, torch.nn.BatchNorm2d) or isinstance(m, torch.nn.BatchNorm1d):
            m.momentum = momentum

    LEARNING_RATE_CLIP = 1e-7
    MOMENTUM_ORIGINAL = 0.1
    MOMENTUM_DECCAY = 0.5
    MOMENTUM_DECCAY_STEP = args.step_size

    global_epoch = 0

    def update_lr(optimizer, epoch, args_lr, args_lr_decay, args_step_size):
        lr = max(args_lr * (args_lr_decay ** (epoch // args_step_size)), LEARNING_RATE_CLIP)
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr
        return lr

    best_loss = 999
    # ---------------- Start training --------------------
    for epoch in range(start_epoch, args.epoch):
        '''Train on scenes'''
        log_string('**** Epoch %d (%d/%s) ****' % (global_epoch + 1, epoch + 1, args.epoch))
        lr = update_lr(optimizer, epoch, args.learning_rate, args.lr_decay, args.step_size)
        log_string(f'Learning rate generator: {lr:.7f}')
        momentum = MOMENTUM_ORIGINAL * (MOMENTUM_DECCAY ** (epoch // MOMENTUM_DECCAY_STEP))
        if momentum < 0.01:
            momentum = 0.01
        print('BN momentum updated to: %f' % momentum)
        network = network.apply(lambda x: bn_momentum_adjust(x, momentum))

        loss_sum = 0
        # ---------------- Start batch set training --------------------
        network = network.train()
        bar = tqdm(enumerate(trainDataLoader), total=len(trainDataLoader), smoothing=0.9)
        for i, data in bar:
            group_posrot, group_onehot, group_lab, floor_mask, N, _ = data
            group_posrot = group_posrot.float().cuda()
            group_onehot = group_onehot.float().cuda()
            group_lab = group_lab.long().cuda()
            floor_mask = floor_mask.float().cuda()
            N = N.long().cuda()

            logprob = network(group_posrot, group_onehot, N)
            loss = criterion(logprob)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            loss_sum = loss_sum + loss.item()

        curr_loss = loss_sum / len(trainDataLoader)
        log_string(f'Training mean loss: {curr_loss:.5f}')
        if best_loss > curr_loss:
            best_loss = curr_loss
            logger.info('Save model...')
            savepath = str(checkpoints_dir) + '/best_model.pth'
            log_string('Saving at %s' % savepath)
            state = {
                'epoch': epoch,
                'current_loss': curr_loss,
                'model_state_dict': network.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
            }
            torch.save(state, savepath)
            log_string('Saving model....')
        log_string(f'Best mean generator loss: {best_loss:.5f}')
        if global_epoch == 500 or global_epoch == 750 or global_epoch == 999:
            logger.info('Save model...')
            savepath = str(checkpoints_dir) + f'/best_model_{global_epoch}.pth'
            log_string('Saving at %s' % savepath)
            state = {
                'epoch': epoch,
                'current_loss': curr_loss,
                'model_state_dict': network.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
            }
            torch.save(state, savepath)
            log_string('Saving model....')
        global_epoch += 1


if __name__ == "__main__":
    np.random.seed()
    args = parse_args()
    main(args)
